import logging
import typing

from ... import platforms, state

# TODO fix these imports after refactoring state module
# from ... import models, state

logger = logging.getLogger(__name__)


# TODO: fix the cpustate type label once refactoring state module is complete
def setup_default_libc(
    flat_api: typing.Any,
    libc_func_names: typing.List[str],
    cpustate: typing.Any,
    canonicalize: bool = True,
) -> None:
    """Map some default libc models into the cpu state.

    Uses Ghdira to figure out entry points in PLT for libc fns and arranges for
    those in a user-provided list to be hooked using the default models in
    Smallworld.  Idea is you might not want all of them mapped and so you say
    just these for now.

    Arguments:
        flat_api: this is what gets returned by pyhidra.open_program(elf_file)
        libc_func_names: list of names of libc functions
        cpustate: cpu state into which to map models
    """

    platform = platforms.Platform(
        platforms.Architecture.X86_64, platforms.Byteorder.LITTLE
    )
    abi = platforms.ABI.SYSTEMV

    program = flat_api.getCurrentProgram()
    listing = program.getListing()

    # find plt section
    plt = None
    for block in program.getMemory().getBlocks():
        if "plt" in block.getName():
            plt = block

    assert plt is not None

    # map all requested libc default models
    num_mapped = 0
    num_no_model = 0
    num_too_many_models = 0
    for func in listing.getFunctions(True):
        func_name = func.getName()
        entry = func.getEntryPoint()
        if not plt.contains(entry):
            continue
        else:
            if func_name in libc_func_names:  # type: ignore
                try:
                    # func is in plt and it is a function for which we want to use a default model
                    int_entry = int(entry.getOffset())
                    model = state.models.Model.lookup(
                        func_name, platform, abi, int_entry
                    )
                    cpustate.map(model)
                    logger.debug(
                        f"Added libc model for {func_name} entry {int_entry:x}"
                    )
                    num_mapped += 1
                except ValueError:
                    logger.debug(
                        f"As there is no default model for {func_name}, adding with null model, entry {int_entry:x}"
                    )
                    model = state.models.Model.lookup("null", platform, abi, int_entry)
                    cpustate.map(model)
                    num_no_model += 1

    logger.info(
        f"Libc model mappings: {num_mapped} default, {num_no_model} no model, {num_too_many_models} too many models"
    )


# TODO: fix the cpustate type label once refactoring state module is complete
def setup_section(
    flat_api: typing.Any,
    section_name: str,
    cpustate: typing.Any,
    elf_file: str = "None",
) -> bytes:
    """Set up this section in cpustate, possibly using contents of elf file

    Uses ghidra to get start addr / size of sections for adding them to
    cpustate. If elf_file is specified, the data will come from the file (ghidra
    tells us where to find it) and get mapped into cpustate memory. Else that
    will be zeros.

    Arguments:
        flat_api: this is what gets returned by pyhidra.open_program(elf_file)
        section_name: '.data' or '.txt' or '.got' or ..
        cpustate: cpu state into which to map models
        elf_file: name of file flat_api came from (elf)

    Returns:
        If elf_file is specified then cpu state for that came from
        file which means we loaded the data out of the file at the
        correct offset.  This will be returned in that case. Else, the
        section data will be 0s.

    """
    # note, elf_file assumed to be same as one flat_api opened
    program = flat_api.getCurrentProgram()
    memory = program.getMemory()
    block = memory.getBlock(section_name)
    address = int(block.start.getOffset())
    size = int(block.size)
    section_bytes = None
    if elf_file == "None":
        # assume we are to just zero this
        section_bytes = b"\0" * size
    else:
        # this is file offset that block
        offs_in_elf = block.getSourceInfos()[0].getFileBytesOffset()
        # read actual section bytesout of elf
        with open(elf_file, "rb") as e:
            e.seek(offs_in_elf)
            section_bytes = e.read(size)
    section = state.memory.RawMemory.from_bytes(section_bytes, address)
    cpustate.map(section, section_name)
    return section_bytes



from angr.storage.memory_mixins.memory_mixin import MemoryMixin


class BaseMemoryMixin(MemoryMixin):  # type: ignore[misc]
    """Base class for memory mixins.

    I wanted to add the _setup_tui() method,
    and needed a base class to make multiple inheritance happen cleanly.
    """

    def _setup_tui(self):
        pass



import copy
import hashlib
import logging
import typing
from enum import Enum

import capstone

import smallworld
from smallworld.analyses.trace_execution_types import CmpInfo, TraceElement, TraceRes
from smallworld.instructions import RegisterOperand

from .. import platforms
from ..hinting.hints import TraceExecutionHint
from . import analysis

logger = logging.getLogger(__name__)


class TraceExecutionCBPoint(Enum):
    BEFORE_INSTRUCTION = 1
    AFTER_INSTRUCTION = 2


def get_cmp_info(
    platform: smallworld.platforms.Platform,
    emulator: smallworld.emulators.Emulator,
    cs_insn: capstone.CsInsn,
) -> typing.Tuple[typing.List[CmpInfo], typing.List[int]]:
    pdefs = platforms.defs.PlatformDef.for_platform(platform)
    if cs_insn.mnemonic in pdefs.compare_mnemonics:
        # it's a compare -- return list of "reads'
        sw_insn = smallworld.instructions.Instruction.from_capstone(cs_insn)
        cmp_info = []
        for op in cs_insn.operands:
            if op.type == capstone.CS_OP_MEM and (op.access & capstone.CS_AC_READ):
                cmp_info.append(sw_insn._memory_reference_operand(op))
            if op.type == capstone.CS_OP_REG and (op.access & capstone.CS_AC_READ):
                cmp_info.append(RegisterOperand(cs_insn.reg_name(op.value.reg)))
            if op.type == capstone.CS_OP_IMM:
                cmp_info.append(op.value.imm)
        immediates = []
        for op in cs_insn.operands:
            if op.type == capstone.x86.X86_OP_IMM:
                immediates.append(op.value.imm)
        return (cmp_info, immediates)
    return ([], [])


class TraceExecution(analysis.Analysis):
    name = "trace_execution"
    description = "perform one concrete execution given a machine state, collecting trace, coverage, and errors"
    version = "0.0.1"

    def __init__(
        self,
        *args,
        num_insns: int,
        seed: int = 1234567,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.num_insns = num_insns
        self.seed = seed
        self.before_instruction_cbs: typing.List[typing.Any] = []
        self.after_instruction_cbs: typing.List[typing.Any] = []

    def register_emu_summary(self):
        pdefs = platforms.defs.PlatformDef.for_platform(self.platform)
        m = hashlib.md5()
        for reg in pdefs.general_purpose_registers:
            vs = (str(self.emulator.read_register(reg))).encode("utf-8")
            m.update(vs)
        return m.hexdigest()

    def register_cb(self, cb_point, cb_function):
        assert isinstance(cb_point, TraceExecutionCBPoint)
        if cb_point == TraceExecutionCBPoint.BEFORE_INSTRUCTION:
            self.before_instruction_cbs.append(cb_function)
        if cb_point == TraceExecutionCBPoint.AFTER_INSTRUCTION:
            self.after_instruction_cbs.append(cb_function)

    def run(self, machine: smallworld.state.Machine) -> None:
        self.machine = copy.deepcopy(machine)
        self.platform = machine.get_platform()
        self.emulator = smallworld.emulators.unicorn.UnicornEmulator(self.platform)
        self.machine.apply(self.emulator)

        logger.debug(f"starting regs in emu {self.register_emu_summary()}")

        def get_insn(pc):
            code = self.emulator.read_memory(pc, 15)
            if code is None:
                raise smallworld.exceptions.AnalysisRunError(
                    "Unable to read next instruction out of emulator memory"
                )
            (cs_insns, disas) = self.emulator._disassemble(code, pc, 1)
            return cs_insns[0]

        the_exc = None
        emu_result = TraceRes.ER_NONE

        pdefs = platforms.defs.PlatformDef.for_platform(self.platform)

        i = 0
        trace = []
        while True:
            pc = self.emulator.read_register("pc")
            cs_insn = get_insn(pc)
            (cmp_info, imm_info) = get_cmp_info(self.platform, self.emulator, cs_insn)
            branch_info = cs_insn.mnemonic in pdefs.conditional_branch_mnemonics
            te = TraceElement(
                pc, i, cs_insn.mnemonic, cs_insn.op_str, cmp_info, branch_info, imm_info
            )
            trace.append(te)
            # run any callbacks
            for before_cb in self.before_instruction_cbs:
                before_cb(self.emulator, pc, te)
            try:
                i += 1
                self.emulator.step()
            except smallworld.exceptions.EmulationBounds:
                # this one really isnt an error of any kind; we
                # encountered code we were not supposed to execute
                emu_result = TraceRes.ER_BOUNDS
                break
            except Exception as e:
                # grab the exception and save it for hinting
                emu_result = TraceRes.ER_FAIL
                the_exc = e
                break
            # run any after callbacks
            for after_cb in self.after_instruction_cbs:
                after_cb(self.emulator, pc, te)
            if i == self.num_insns:
                emu_result = TraceRes.ER_MAX_INSNS
                break

        m = hashlib.md5()
        for te in trace:
            logger.debug(te)
            m.update((str(te.pc).encode("utf-8")))

        logger.info(
            f"captured trace of {i} instructions, res={emu_result} trace_digest={m.hexdigest()}"
        )

        assert emu_result is not None

        hint = TraceExecutionHint(
            message="A single execution trace",
            trace=trace,
            trace_digest=m.hexdigest(),
            seed=self.seed,
            emu_result=emu_result,
            exception=the_exc,
            exception_class=str(type(the_exc)),
        )

        self.hinter.send(hint)



import typing

from .... import emulators
from ..cstd import ArgumentType, CStdModel

# NOTE: Most of signal.h is not defined in C99
#
# If you're familiar with Linux,
# the rest of the functions you're looking for
# are part of the POSIX standard.
#
# Windows only implements the C99 API;
# I suspect it keeps most process lifecycle controls
# internal to its own application runtimes.


class Raise(CStdModel):
    name = "raise"

    # int raise(int sig);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.INT

    # Most of our emulators can't model signals.
    # I'm not sure it's possible to model raising one.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Signal(CStdModel):
    name = "signal"

    # typedef void (*sighandler_t)(int);
    # sighandler_t signal(int sig, sighandler_t func);
    argument_types = [ArgumentType.INT, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    # Most of our emulators can't model signals.
    imprecise = True

    def __init__(self, address: int):
        super().__init__(address)
        # Track any registered handlers.
        # Sometimes programs care about this.
        self.handlers: typing.Dict[int, int] = dict()

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        sig = self.get_arg1(emulator)
        func = self.get_arg2(emulator)

        assert isinstance(sig, int)
        assert isinstance(func, int)

        # Get the old handler.  Default to SIG_DFL, or NULL.
        old_func = self.handlers.get(sig, 0)
        self.handlers[sig] = func

        self.set_return_value(emulator, old_func)


__all__ = [
    "Raise",
    "Signal",
]



import logging
import typing

import claripy
import jpype
from ghidra.pcode.emu import PcodeEmulator, PcodeThread
from ghidra.pcode.exec import PcodeExecutorStatePiece
from ghidra.program.model.address import AddressRangeImpl
from ghidra.program.model.pcode import Varnode

from ... import exceptions, platforms, utils
from ..emulator import Emulator
from .machdefs import GhidraMachineDef
from .typing import AbstractGhidraEmulator

log = logging.getLogger(__name__)


class GhidraEmulator(AbstractGhidraEmulator):
    name = "pcode-emulator"
    description = "Emulator based on pyghidra and pcode"
    version = "0.0.1"

    # Convert bytes into a JPype byte[]
    # bytes and bytearray are considered variants of String by JPype
    bytes_py_to_java = jpype.JByte[:]

    @staticmethod
    def bytes_java_to_py(val: jpype.JByte[:]):
        # Convert a JPype byte[] into a bytes object
        bytelist = list(
            map(lambda x: x.numerator if x.numerator >= 0 else 256 + x.numerator, val)
        )
        return bytes(bytelist)

    def __init__(self, platform: platforms.Platform):
        super().__init__(platform)
        self.platform: platforms.Platform = platform
        self.platdef: platforms.PlatformDef = platforms.PlatformDef.for_platform(
            platform
        )
        self.machdef: GhidraMachineDef = GhidraMachineDef.for_platform(platform)

        self._emu: PcodeEmulator = PcodeEmulator(self.machdef.language)
        # Set up the context configuration.
        # This includes execution mode information,
        # and isn't automatically propagated to the thread.
        self._thread.overrideContextWithDefault()

        self._memory_map = utils.RangeCollection()

        # Instruction hooking callbacks
        self._instructions_hook: typing.Optional[
            typing.Callable[[Emulator], None]
        ] = None
        self._instruction_hooks: typing.Dict[
            int, typing.Callable[[Emulator], None]
        ] = dict()

        # Function hooking callbacks
        self._function_hooks: typing.Dict[
            int, typing.Callable[[Emulator], None]
        ] = dict()

        # Memory hooking callbacks
        self._mem_reads_hook: typing.Optional[
            typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]]
        ] = None
        self._mem_read_hooks: typing.Dict[
            typing.Tuple[int, int],
            typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]],
        ] = dict()
        self._mem_writes_hook: typing.Optional[
            typing.Callable[[Emulator, int, int, bytes], None]
        ] = None
        self._mem_write_hooks: typing.Dict[
            typing.Tuple[int, int], typing.Callable[[Emulator, int, int, bytes], None]
        ] = dict()

    @property
    def _thread(self) -> PcodeThread:
        return self._emu.getThread("main", True)

    def read_register_content(self, name: str) -> int:
        # Determine address and size of register
        if name == "pc":
            name = self.platdef.pc_register
        reg = self.machdef.pcode_reg(name)

        # Get the thread's register state
        state = self._thread.getState()

        # Get the register value in bytes
        val = state.getVar(reg, PcodeExecutorStatePiece.Reason.INSPECT)

        if self.platform.byteorder is platforms.Byteorder.LITTLE:
            return int.from_bytes(val, "little")
        elif self.platform.byteorder is platforms.Byteorder.BIG:
            return int.from_bytes(val, "big")
        else:
            raise Exception("Unable to decode byteorder {self.platform.byteorder}")

    def write_register_content(
        self, name: str, value: typing.Union[None, int, claripy.ast.bv.BV]
    ) -> None:
        # Determine address and size of register
        if name == "pc":
            name = self.platdef.pc_register
        reg = self.machdef.pcode_reg(name)

        if value is None:
            return
        if not isinstance(value, int):
            raise TypeError("Pcode emulator can't support symbolic values")

        state = self._thread.getState()

        if self.platform.byteorder is platforms.Byteorder.BIG:
            val = value.to_bytes(reg.getMinimumByteSize(), "big")
        elif self.platform.byteorder is platforms.Byteorder.LITTLE:
            val = value.to_bytes(reg.getMinimumByteSize(), "little")
        else:
            raise Exception("Unable to encode byteorder {self.platform.byteorder}")

        state.setVar(reg, self.bytes_py_to_java(val))

    def read_memory_content(self, address: int, size: int) -> bytes:
        # Get the thread's memory state
        shared = self._emu.getSharedState()

        if self.platform.byteorder is platforms.Byteorder.BIG:
            addr_bytes = address.to_bytes(self.platdef.address_size, "big")
        elif self.platform.byteorder is platforms.Byteorder.LITTLE:
            addr_bytes = address.to_bytes(self.platdef.address_size, "little")
        else:
            raise Exception("Unable to encode byteorder {self.platform.byteorder}")

        # Get the data out of the default address space
        # NOTE: Ghidra can support machines with multiple address spaces.
        # SmallWorld does not.
        val = shared.getVar(
            self.machdef.language.getDefaultSpace(),
            self.bytes_py_to_java(addr_bytes),
            size,
            False,
            shared.Reason.INSPECT,
        )
        return self.bytes_java_to_py(val)

    def map_memory(self, address: int, size: int) -> None:
        region = (address, address + size)
        self._memory_map.add_range(region)

    def get_memory_map(self) -> typing.List[typing.Tuple[int, int]]:
        return list(self._memory_map.ranges)

    def write_memory_content(
        self, address: int, content: typing.Union[bytes, claripy.ast.bv.BV]
    ):
        if isinstance(content, claripy.ast.bv.BV):
            raise TypeError("Pcode emulator can't handle symbolic expressions")

        log.info(f"Writing {hex(len(content))} bytes at {hex(address)}")

        # Get the thread's memory state
        shared = self._emu.getSharedState()

        val = self.bytes_py_to_java(content)
        if self.platform.byteorder is platforms.Byteorder.BIG:
            addr_bytes = address.to_bytes(self.platdef.address_size, "big")
        elif self.platform.byteorder is platforms.Byteorder.LITTLE:
            addr_bytes = address.to_bytes(self.platdef.address_size, "little")
        else:
            raise Exception("Unable to encode byteorder {self.platform.byteorder}")

        # Get the data out of the default address space
        # NOTE: Ghidra can support machines with multiple address spaces.
        # SmallWorld does not.
        shared.setVar(
            self.machdef.language.getDefaultSpace(),
            self.bytes_py_to_java(addr_bytes),
            len(content),
            False,
            val,
        )

    def hook_instruction(
        self, address: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        self._instruction_hooks[address] = function

    def unhook_instruction(self, address: int) -> None:
        if address in self._instruction_hooks:
            del self._instruction_hooks[address]

    def hook_instructions(self, function: typing.Callable[[Emulator], None]) -> None:
        self._instructions_hook = function

    def unhook_instructions(self) -> None:
        self._instructions_hook = None

    def _process_function_hook(self, address: int):
        if address not in self._function_hooks:
            # This really should never happen unless someone calls this manually.
            raise exceptions.ConfigurationError(f"No function hook at {hex(address)}")

        # Run the hook
        self._function_hooks[address](self)

        # Mimic a platform-specific "return" instruction.
        if self.platform.architecture == platforms.Architecture.X86_32:
            sp = self.read_register("esp")
            if self.platform.byteorder == platforms.Byteorder.LITTLE:
                ret = int.from_bytes(self.read_memory(sp, 4), "little")
            elif self.platform.byteorder == platforms.Byteorder.BIG:
                ret = int.from_bytes(self.read_memory(sp, 4), "big")
            self.write_register("esp", sp + 4)
        elif self.platform.architecture == platforms.Architecture.X86_64:
            # amd64: pop an 8-byte value off the stack
            sp = self.read_register("rsp")
            if self.platform.byteorder == platforms.Byteorder.LITTLE:
                ret = int.from_bytes(self.read_memory(sp, 8), "little")
            elif self.platform.byteorder == platforms.Byteorder.BIG:
                ret = int.from_bytes(self.read_memory(sp, 8), "big")
            self.write_register("rsp", sp + 8)
        elif (
            self.platform.architecture == platforms.Architecture.AARCH64
            or self.platform.architecture == platforms.Architecture.ARM_V5T
            or self.platform.architecture == platforms.Architecture.ARM_V6M
            or self.platform.architecture == platforms.Architecture.ARM_V6M_THUMB
            or self.platform.architecture == platforms.Architecture.ARM_V7A
            or self.platform.architecture == platforms.Architecture.ARM_V7M
            or self.platform.architecture == platforms.Architecture.ARM_V7R
            or self.platform.architecture == platforms.Architecture.POWERPC32
            or self.platform.architecture == platforms.Architecture.POWERPC64
        ):
            # aarch64, arm32, powerpc and powerpc64: branch to register 'lr'
            ret = self.read_register("lr")
        elif (
            self.platform.architecture == platforms.Architecture.LOONGARCH64
            or self.platform.architecture == platforms.Architecture.MIPS32
            or self.platform.architecture == platforms.Architecture.MIPS64
            or self.platform.architecture == platforms.Architecture.RISCV64
        ):
            # mips32, mips64, and riscv64: branch to register 'ra'
            ret = self.read_register("ra")
        elif self.platform.architecture == platforms.Architecture.XTENSA:
            # xtensa: branch to register 'a0'
            ret = self.read_register("a0")

        self.write_register("pc", ret)

    def hook_function(
        self, address: int, function: typing.Callable[[Emulator], None]
    ) -> None:
        self._function_hooks[address] = function

    def unhook_function(self, address: int) -> None:
        if address in self._function_hooks:
            del self._function_hooks[address]

    def _update_access_breakpoints(self) -> None:
        # Refresh all access breakpoints
        # There's no way to clear a single access breakpoint;
        # you need to clear them all and re-apply.
        return

        # Wipe out all access breakpoints
        self._emu.clearAccessBreakpoints()

        addrspace = self.machdef.language.getDefaultSpace()

        # Add back the global read breakpoint
        if self._mem_reads_hook is not None:
            addr_range = AddressRangeImpl(
                addrspace.getMinAddress(), addrspace.getMaxAddress()
            )
            self._emu.addAccessBreakpoint(addr_range, self._emu.AccessKind.R)

        # Add back the specific read breakpoints
        for start, end in self._mem_read_hooks:
            start_addr = addrspace.getAddress(start)
            end_addr = addrspace.getAddress(end)
            addr_range = AddressRangeImpl(start_addr, end_addr)
            self._emu.addAccessBreakpoint(addr_range, self._emu.AccessKind.R)

        # Add back the global write breakpoint
        if self._mem_writes_hook is not None:
            addr_range = AddressRangeImpl(
                addrspace.getMinAddress(), addrspace.getMaxAddress()
            )
            self._emu.addAccessBreakpoint(addr_range, self._emu.AccessKind.W)

        # Add back the specific write breakpoints
        for start, end in self._mem_write_hooks:
            start_addr = addrspace.getAddress(start)
            end_addr = addrspace.getAddress(end)
            addr_range = AddressRangeImpl(start_addr, end_addr)
            self._emu.addAccessBreakpoint(addr_range, self._emu.AccessKind.W)

    def _process_read_breakpoint(self, addr_var: Varnode, out_var: Varnode) -> None:
        state = self._thread.getState()

        # Get the address out of the emulator
        addr_bytes = state.getVar(addr_var, PcodeExecutorStatePiece.Reason.INSPECT)
        if self.platform.byteorder is platforms.Byteorder.BIG:
            addr = int.from_bytes(addr_bytes, "big")
        elif self.platform.byteorder is platforms.Byteorder.LITTLE:
            addr = int.from_bytes(addr_bytes, "little")

        # Dereference the address to get the original data
        addr_space = self.machdef.language.getDefaultSpace()
        addr_addr = addr_space.getAddress(addr)
        data_var = Varnode(addr_addr, out_var.getSize())
        data = state.getVar(data_var, PcodeExecutorStatePiece.Reason.INSPECT)

        if self._mem_reads_hook is not None:
            new_data = self._mem_reads_hook(self, addr, len(data), data)
            if new_data is not None:
                data = new_data

        end_addr = addr + len(data)

        for (start, end), hook in self._mem_read_hooks.items():
            rng = range(start, end)
            access_rng = range(addr, end_addr)
            if (
                addr in rng
                or end_addr - 1 in rng
                or start in access_rng
                or end - 1 in access_rng
            ):
                new_data = hook(self, addr, len(data), data)
                if new_data is not None:
                    data = new_data

        # Write the data to the output
        #
        # CRITICAL: This must always happen, even if the op wasn't hooked!
        # The main processing loop skips any LOAD opcodes,
        # since their normal execution will clobber any value produced
        # by this handler.
        # Thus, this handler must fully replace the behavior of LOAD.
        state.setVar(out_var, self.bytes_py_to_java(data))

    def hook_memory_read(
        self,
        start: int,
        end: int,
        function: typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]],
    ) -> None:
        self._mem_read_hooks[(start, end)] = function
        self._update_access_breakpoints()

    def unhook_memory_read(self, start: int, end: int) -> None:
        if (start, end) in self._mem_read_hooks:
            del self._mem_read_hooks[(start, end)]
            self._update_access_breakpoints()

    def hook_memory_reads(
        self,
        function: typing.Callable[[Emulator, int, int, bytes], typing.Optional[bytes]],
    ) -> None:
        self._mem_reads_hook = function
        self._update_access_breakpoints()

    def unhook_memory_reads(self) -> None:
        if self._mem_reads_hook is not None:
            self._mem_reads_hook = None
            self._update_access_breakpoints()

    def _process_write_breakpoint(self, addr_var: Varnode, data_var: Varnode) -> None:
        state = self._thread.getState()
        addr_bytes = state.getVar(addr_var, PcodeExecutorStatePiece.Reason.INSPECT)
        data = state.getVar(data_var, PcodeExecutorStatePiece.Reason.INSPECT)

        if self.platform.byteorder is platforms.Byteorder.BIG:
            addr = int.from_bytes(addr_bytes, "big")
        elif self.platform.byteorder is platforms.Byteorder.LITTLE:
            addr = int.from_bytes(addr_bytes, "little")

        if self._mem_writes_hook is not None:
            self._mem_writes_hook(self, addr, len(data), data)

        end_addr = addr + len(data)

        for (start, end), hook in self._mem_write_hooks.items():
            rng = range(start, end)
            access_rng = range(addr, end_addr)
            if (
                addr in rng
                or end_addr - 1 in rng
                or start in access_rng
                or end - 1 in access_rng
            ):
                hook(self, addr, len(data), data)

    def hook_memory_write(
        self,
        start: int,
        end: int,
        function: typing.Callable[[Emulator, int, int, bytes], None],
    ) -> None:
        self._mem_write_hooks[(start, end)] = function
        self._update_access_breakpoints()

    def unhook_memory_write(self, start: int, end: int) -> None:
        if (start, end) in self._mem_write_hooks:
            del self._mem_write_hooks[(start, end)]
            self._update_access_breakpoints()

    def hook_memory_writes(
        self,
        function: typing.Callable[[Emulator, int, int, bytes], None],
    ) -> None:
        self._mem_writes_hook = function
        self._update_access_breakpoints()

    def unhook_memory_writes(self) -> None:
        if self._mem_writes_hook is not None:
            self._mem_writes_hook = None
            self._update_access_breakpoints()

    def step_instruction(self) -> None:
        if not self.machdef.supports_single_step:
            raise exceptions.ConfigurationError(
                f"PcodeEmulator does not support single-instruction stepping for {self.platform}"
            )

        # Step!
        pc = self.read_register_content(self.platdef.pc_register)
        pc_addr = self.machdef.language.getDefaultSpace().getAddress(pc)
        self._thread.overrideCounter(pc_addr)

        # Check for instruction hooks.
        # TODO: Should instruction hooks override the instruction?
        if self._instructions_hook is not None:
            self._instructions_hook(self)
        if pc in self._instruction_hooks:
            self._instruction_hooks[pc](self)

        # Check for function hooks
        if pc in self._function_hooks:
            # Function hooks bypass normal instruction processing,
            # but I still want bounds/exits to work normally.
            self._process_function_hook(pc)
        else:
            skip = False
            frame = None
            # We need to step individual pcode ops to capture hooks
            # There is an exception-based hooking mechanism,
            # but it's unworkable for SmallWorld's purpose.
            while True:
                if skip:
                    # Skip the next op; it was modeled by a hook
                    skip = False
                    self._thread.skipPcodeOp()
                else:
                    # Execute the opcode normally
                    self._thread.stepPcodeOp()

                frame = self._thread.getFrame()

                if frame is None:
                    # No frame; this is the end of the instruction
                    break

                if frame.isFinished():
                    # Frame is finished; this is the end of the instruction
                    self._thread.finishInstruction()
                    break

                # Inspect the op to see if it's memory-hook relevant
                code = frame.getCode()
                op = code[frame.index()]
                opcode = op.getOpcode()
                if opcode == op.STORE:
                    # This is a STORE opcode; could trigger a "write" hook
                    _, addr_var, data_var = op.getInputs()
                    self._process_write_breakpoint(addr_var, data_var)
                elif opcode == op.LOAD:
                    # This is a LOAD opcode; could trigger a "read" hook
                    _, addr_var = op.getInputs()
                    out_var = op.getOutput()
                    self._process_read_breakpoint(addr_var, out_var)
                    # Skip the actual LOAD opcode
                    # The read breakpoint handler will mimic its behavior;
                    # running the op normally will clobber a custom value
                    # produced by a hook.
                    skip = True

        # Check exit points and bounds
        pc = self.read_register_content(self.platdef.pc_register)

        if pc in self._exit_points:
            raise exceptions.EmulationExitpoint()
        if not self._bounds.is_empty() and not self._bounds.contains_value(pc):
            raise exceptions.EmulationBounds()
        if not self._memory_map.is_empty() and not self._memory_map.contains_value(pc):
            raise exceptions.EmulationBounds()

    def step_block(self) -> None:
        raise NotImplementedError("Not sure how to step by block.")

    def run(self) -> None:
        try:
            while True:
                self.step_instruction()
        except exceptions.EmulationStop:
            pass

    def __repr__(self) -> str:
        return "PcodeEmulator"



from ... import platforms
from .. import state
from . import cpu


class LoongArch(cpu.CPU):
    """Abstract CPU state object for all LoongArch targets"""

    def __init__(self, wordsize: int) -> None:
        super().__init__()
        # Program counter (not a real register)
        self.pc = state.Register("pc", size=wordsize)
        self.add(self.pc)
        # Zero register
        self.r0 = state.Register("r0", size=wordsize)
        self.add(self.r0)
        self.zero = state.RegisterAlias("zero", self.r0, size=wordsize, offset=0)
        self.add(self.zero)
        # Return address
        self.r1 = state.Register("r1", size=wordsize)
        self.add(self.r1)
        self.ra = state.RegisterAlias("ra", self.r1, size=wordsize, offset=0)
        self.add(self.ra)
        # TLS pointer
        self.r2 = state.Register("r2", size=wordsize)
        self.add(self.r2)
        self.tp = state.RegisterAlias("tp", self.r2, size=wordsize, offset=0)
        self.add(self.tp)
        # Stack pointer
        self.r3 = state.Register("r3", size=wordsize)
        self.add(self.r3)
        self.sp = state.RegisterAlias("sp", self.r3, size=wordsize, offset=0)
        self.add(self.sp)
        # Arguments.
        # a0 and a1 are also the return registers
        self.r4 = state.Register("r4", size=wordsize)
        self.add(self.r4)
        self.a0 = state.RegisterAlias("a0", self.r4, size=wordsize, offset=0)
        self.add(self.a0)
        self.v0 = state.RegisterAlias("v0", self.r4, size=wordsize, offset=0)
        self.add(self.v0)
        self.r5 = state.Register("r5", size=wordsize)
        self.add(self.r5)
        self.a1 = state.RegisterAlias("a1", self.r5, size=wordsize, offset=0)
        self.add(self.a1)
        self.v1 = state.RegisterAlias("v1", self.r5, size=wordsize, offset=0)
        self.add(self.v1)
        self.r6 = state.Register("r6", size=wordsize)
        self.add(self.r6)
        self.a2 = state.RegisterAlias("a2", self.r6, size=wordsize, offset=0)
        self.add(self.a2)
        self.r7 = state.Register("r7", size=wordsize)
        self.add(self.r7)
        self.a3 = state.RegisterAlias("a3", self.r7, size=wordsize, offset=0)
        self.add(self.a3)
        self.r8 = state.Register("r8", size=wordsize)
        self.add(self.r8)
        self.a4 = state.RegisterAlias("a4", self.r8, size=wordsize, offset=0)
        self.add(self.a4)
        self.r9 = state.Register("r9", size=wordsize)
        self.add(self.r9)
        self.a5 = state.RegisterAlias("a5", self.r9, size=wordsize, offset=0)
        self.add(self.a5)
        self.r10 = state.Register("r10", size=wordsize)
        self.add(self.r10)
        self.a6 = state.RegisterAlias("a6", self.r10, size=wordsize, offset=0)
        self.add(self.a6)
        self.r11 = state.Register("r11", size=wordsize)
        self.add(self.r11)
        self.a7 = state.RegisterAlias("a7", self.r11, size=wordsize, offset=0)
        self.add(self.a7)
        # Temporary registers
        self.r12 = state.Register("r12", size=wordsize)
        self.add(self.r12)
        self.t0 = state.RegisterAlias("t0", self.r12, size=wordsize, offset=0)
        self.add(self.t0)
        self.r13 = state.Register("r13", size=wordsize)
        self.add(self.r13)
        self.t1 = state.RegisterAlias("t1", self.r13, size=wordsize, offset=0)
        self.add(self.t1)
        self.r14 = state.Register("r14", size=wordsize)
        self.add(self.r14)
        self.t2 = state.RegisterAlias("t2", self.r14, size=wordsize, offset=0)
        self.add(self.t2)
        self.r15 = state.Register("r15", size=wordsize)
        self.add(self.r15)
        self.t3 = state.RegisterAlias("t3", self.r15, size=wordsize, offset=0)
        self.add(self.t3)
        self.r16 = state.Register("r16", size=wordsize)
        self.add(self.r16)
        self.t4 = state.RegisterAlias("t4", self.r16, size=wordsize, offset=0)
        self.add(self.t4)
        self.r17 = state.Register("r17", size=wordsize)
        self.add(self.r17)
        self.t5 = state.RegisterAlias("t5", self.r17, size=wordsize, offset=0)
        self.add(self.t5)
        self.r18 = state.Register("r18", size=wordsize)
        self.add(self.r18)
        self.t6 = state.RegisterAlias("t6", self.r18, size=wordsize, offset=0)
        self.add(self.t6)
        self.r19 = state.Register("r19", size=wordsize)
        self.add(self.r19)
        self.t7 = state.RegisterAlias("t7", self.r19, size=wordsize, offset=0)
        self.add(self.t7)
        self.r20 = state.Register("r20", size=wordsize)
        self.add(self.r20)
        self.t8 = state.RegisterAlias("t8", self.r20, size=wordsize, offset=0)
        self.add(self.t8)
        # Per-CPU Base Address
        self.r21 = state.Register("r21", size=wordsize)
        self.add(self.r21)
        self.u0 = state.RegisterAlias("u0", self.r21, size=wordsize, offset=0)
        self.add(self.u0)
        # Frame Pointer
        self.r22 = state.Register("r22", size=wordsize)
        self.add(self.r22)
        self.fp = state.RegisterAlias("fp", self.r22, size=wordsize, offset=0)
        self.add(self.fp)
        # Static registers
        self.r23 = state.Register("r23", size=wordsize)
        self.add(self.r23)
        self.s0 = state.RegisterAlias("s0", self.r23, size=wordsize, offset=0)
        self.add(self.s0)
        self.r24 = state.Register("r24", size=wordsize)
        self.add(self.r24)
        self.s1 = state.RegisterAlias("s1", self.r24, size=wordsize, offset=0)
        self.add(self.s1)
        self.r25 = state.Register("r25", size=wordsize)
        self.add(self.r25)
        self.s2 = state.RegisterAlias("s2", self.r25, size=wordsize, offset=0)
        self.add(self.s2)
        self.r26 = state.Register("r26", size=wordsize)
        self.add(self.r26)
        self.s3 = state.RegisterAlias("s3", self.r26, size=wordsize, offset=0)
        self.add(self.s3)
        self.r27 = state.Register("r27", size=wordsize)
        self.add(self.r27)
        self.s4 = state.RegisterAlias("s4", self.r27, size=wordsize, offset=0)
        self.add(self.s4)
        self.r28 = state.Register("r28", size=wordsize)
        self.add(self.r28)
        self.s5 = state.RegisterAlias("s5", self.r28, size=wordsize, offset=0)
        self.add(self.s5)
        self.r29 = state.Register("r29", size=wordsize)
        self.add(self.r29)
        self.s6 = state.RegisterAlias("s6", self.r29, size=wordsize, offset=0)
        self.add(self.s6)
        self.r30 = state.Register("r30", size=wordsize)
        self.add(self.r30)
        self.s7 = state.RegisterAlias("s7", self.r30, size=wordsize, offset=0)
        self.add(self.s7)
        self.r31 = state.Register("r31", size=wordsize)
        self.add(self.r31)
        self.s8 = state.RegisterAlias("s8", self.r31, size=wordsize, offset=0)
        self.add(self.s8)
        # Floating-point arguments.
        # fa0 and fa1 are also return values
        self.f0 = state.Register("f0", size=8)
        self.add(self.f0)
        self.fa0 = state.RegisterAlias("fa0", self.f0, size=8, offset=0)
        self.add(self.fa0)
        self.f1 = state.Register("f1", size=8)
        self.add(self.f1)
        self.fa1 = state.RegisterAlias("fa1", self.f1, size=8, offset=0)
        self.add(self.fa1)
        self.f2 = state.Register("f2", size=8)
        self.add(self.f2)
        self.fa2 = state.RegisterAlias("fa2", self.f2, size=8, offset=0)
        self.add(self.fa2)
        self.f3 = state.Register("f3", size=8)
        self.add(self.f3)
        self.fa3 = state.RegisterAlias("fa3", self.f3, size=8, offset=0)
        self.add(self.fa3)
        self.f4 = state.Register("f4", size=8)
        self.add(self.f4)
        self.fa4 = state.RegisterAlias("fa4", self.f4, size=8, offset=0)
        self.add(self.fa4)
        self.f5 = state.Register("f5", size=8)
        self.add(self.f5)
        self.fa5 = state.RegisterAlias("fa5", self.f5, size=8, offset=0)
        self.add(self.fa5)
        self.f6 = state.Register("f6", size=8)
        self.add(self.f6)
        self.fa6 = state.RegisterAlias("fa6", self.f6, size=8, offset=0)
        self.add(self.fa6)
        self.f7 = state.Register("f7", size=8)
        self.add(self.f7)
        self.fa7 = state.RegisterAlias("fa7", self.f7, size=8, offset=0)
        self.add(self.fa7)
        # Floating-point temporary registers
        self.f8 = state.Register("f8", size=8)
        self.add(self.f8)
        self.ft0 = state.RegisterAlias("ft0", self.f8, size=8, offset=0)
        self.add(self.ft0)
        self.f9 = state.Register("f9", size=8)
        self.add(self.f9)
        self.ft1 = state.RegisterAlias("ft1", self.f9, size=8, offset=0)
        self.add(self.ft1)
        self.f10 = state.Register("f10", size=8)
        self.add(self.f10)
        self.ft2 = state.RegisterAlias("ft2", self.f10, size=8, offset=0)
        self.add(self.ft2)
        self.f11 = state.Register("f11", size=8)
        self.add(self.f11)
        self.ft3 = state.RegisterAlias("ft3", self.f11, size=8, offset=0)
        self.add(self.ft3)
        self.f12 = state.Register("f12", size=8)
        self.add(self.f12)
        self.ft4 = state.RegisterAlias("ft4", self.f12, size=8, offset=0)
        self.add(self.ft4)
        self.f13 = state.Register("f13", size=8)
        self.add(self.f13)
        self.ft5 = state.RegisterAlias("ft5", self.f13, size=8, offset=0)
        self.add(self.ft5)
        self.f14 = state.Register("f14", size=8)
        self.add(self.f14)
        self.ft6 = state.RegisterAlias("ft6", self.f14, size=8, offset=0)
        self.add(self.ft6)
        self.f15 = state.Register("f15", size=8)
        self.add(self.f15)
        self.ft7 = state.RegisterAlias("ft7", self.f15, size=8, offset=0)
        self.add(self.ft7)
        self.f16 = state.Register("f16", size=8)
        self.add(self.f16)
        self.ft8 = state.RegisterAlias("ft8", self.f16, size=8, offset=0)
        self.add(self.ft8)
        self.f17 = state.Register("f17", size=8)
        self.add(self.f17)
        self.ft9 = state.RegisterAlias("ft9", self.f17, size=8, offset=0)
        self.add(self.ft9)
        self.f18 = state.Register("f18", size=8)
        self.add(self.f18)
        self.ft10 = state.RegisterAlias("ft10", self.f18, size=8, offset=0)
        self.add(self.ft10)
        self.f19 = state.Register("f19", size=8)
        self.add(self.f19)
        self.ft11 = state.RegisterAlias("ft11", self.f19, size=8, offset=0)
        self.add(self.ft11)
        self.f20 = state.Register("f20", size=8)
        self.add(self.f20)
        self.ft12 = state.RegisterAlias("ft12", self.f20, size=8, offset=0)
        self.add(self.ft12)
        self.f21 = state.Register("f21", size=8)
        self.add(self.f21)
        self.ft13 = state.RegisterAlias("ft13", self.f21, size=8, offset=0)
        self.add(self.ft13)
        self.f22 = state.Register("f22", size=8)
        self.add(self.f22)
        self.ft14 = state.RegisterAlias("ft14", self.f22, size=8, offset=0)
        self.add(self.ft14)
        self.f23 = state.Register("f23", size=8)
        self.add(self.f23)
        self.ft15 = state.RegisterAlias("ft15", self.f23, size=8, offset=0)
        self.add(self.ft15)
        # Floating-point static registers
        self.f24 = state.Register("f24", size=8)
        self.add(self.f24)
        self.fs0 = state.RegisterAlias("fs0", self.f24, size=8, offset=0)
        self.add(self.fs0)
        self.f25 = state.Register("f25", size=8)
        self.add(self.f25)
        self.fs1 = state.RegisterAlias("fs1", self.f25, size=8, offset=0)
        self.add(self.fs1)
        self.f26 = state.Register("f26", size=8)
        self.add(self.f26)
        self.fs2 = state.RegisterAlias("fs2", self.f26, size=8, offset=0)
        self.add(self.fs2)
        self.f27 = state.Register("f27", size=8)
        self.add(self.f27)
        self.fs3 = state.RegisterAlias("fs3", self.f27, size=8, offset=0)
        self.add(self.fs3)
        self.f28 = state.Register("f28", size=8)
        self.add(self.f28)
        self.fs4 = state.RegisterAlias("fs4", self.f28, size=8, offset=0)
        self.add(self.fs4)
        self.f29 = state.Register("f29", size=8)
        self.add(self.f29)
        self.fs5 = state.RegisterAlias("fs5", self.f29, size=8, offset=0)
        self.add(self.fs5)
        self.f30 = state.Register("f30", size=8)
        self.add(self.f30)
        self.fs6 = state.RegisterAlias("fs6", self.f30, size=8, offset=0)
        self.add(self.fs6)
        self.f31 = state.Register("f31", size=8)
        self.add(self.f31)
        self.fs7 = state.RegisterAlias("fs7", self.f31, size=8, offset=0)
        self.add(self.fs7)


class LoongArch64(LoongArch):
    platform = platforms.Platform(
        platforms.Architecture.LOONGARCH64, platforms.Byteorder.LITTLE
    )

    def __init__(self):
        super().__init__(8)


__all__ = ["LoongArch64"]



from .field_analysis import (
    FDAState,
    FieldDetectionAnalysis,
    ForcedFieldDetectionAnalysis,
)
from .malloc import FreeModel, MallocModel

__all__ = [
    "FDAState",
    "FieldDetectionAnalysis",
    "ForcedFieldDetectionAnalysis",
    "FreeModel",
    "MallocModel",
]



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import LoongArch64SysVModel


class LoongArch64SysVAbs(Abs, LoongArch64SysVModel):
    pass


class LoongArch64SysVLAbs(LAbs, LoongArch64SysVModel):
    pass


class LoongArch64SysVLLAbs(LLAbs, LoongArch64SysVModel):
    pass


class LoongArch64SysVAtof(Atof, LoongArch64SysVModel):
    pass


class LoongArch64SysVAtoi(Atoi, LoongArch64SysVModel):
    pass


class LoongArch64SysVAtol(Atol, LoongArch64SysVModel):
    pass


class LoongArch64SysVAtoll(Atoll, LoongArch64SysVModel):
    pass


class LoongArch64SysVCalloc(Calloc, LoongArch64SysVModel):
    pass


class LoongArch64SysVDiv(Div, LoongArch64SysVModel):
    pass


class LoongArch64SysVLDiv(LDiv, LoongArch64SysVModel):
    pass


class LoongArch64SysVLLDiv(LLDiv, LoongArch64SysVModel):
    pass


class LoongArch64SysVExit(Exit, LoongArch64SysVModel):
    pass


class LoongArch64SysVFree(Free, LoongArch64SysVModel):
    pass


class LoongArch64SysVMalloc(Malloc, LoongArch64SysVModel):
    pass


class LoongArch64SysVQSort(QSort, LoongArch64SysVModel):
    pass


class LoongArch64SysVRand(Rand, LoongArch64SysVModel):
    pass


class LoongArch64SysVRealloc(Realloc, LoongArch64SysVModel):
    pass


class LoongArch64SysVSrand(Srand, LoongArch64SysVModel):
    pass


class LoongArch64SysVAbort(Abort, LoongArch64SysVModel):
    pass


class LoongArch64SysVAtexit(Atexit, LoongArch64SysVModel):
    pass


class LoongArch64SysVGetenv(Getenv, LoongArch64SysVModel):
    pass


class LoongArch64SysVMblen(Mblen, LoongArch64SysVModel):
    pass


class LoongArch64SysVMbstowcs(Mbstowcs, LoongArch64SysVModel):
    pass


class LoongArch64SysVMbtowc(Mbtowc, LoongArch64SysVModel):
    pass


class LoongArch64SysVSystem(System, LoongArch64SysVModel):
    pass


class LoongArch64SysVWcstombs(Wcstombs, LoongArch64SysVModel):
    pass


class LoongArch64SysVWctomb(Wctomb, LoongArch64SysVModel):
    pass


__all__ = [
    "LoongArch64SysVAbs",
    "LoongArch64SysVLAbs",
    "LoongArch64SysVLLAbs",
    "LoongArch64SysVAtof",
    "LoongArch64SysVAtoi",
    "LoongArch64SysVAtol",
    "LoongArch64SysVAtoll",
    "LoongArch64SysVCalloc",
    "LoongArch64SysVDiv",
    "LoongArch64SysVLDiv",
    "LoongArch64SysVLLDiv",
    "LoongArch64SysVExit",
    "LoongArch64SysVFree",
    "LoongArch64SysVMalloc",
    "LoongArch64SysVQSort",
    "LoongArch64SysVRand",
    "LoongArch64SysVRealloc",
    "LoongArch64SysVSrand",
    "LoongArch64SysVAbort",
    "LoongArch64SysVAtexit",
    "LoongArch64SysVGetenv",
    "LoongArch64SysVMblen",
    "LoongArch64SysVMbstowcs",
    "LoongArch64SysVMbtowc",
    "LoongArch64SysVSystem",
    "LoongArch64SysVWcstombs",
    "LoongArch64SysVWctomb",
]



from .aarch64 import AArch64Stack
from .amd64 import AMD64Stack
from .arm import ARMv5tStack, ARMv6mStack, ARMv7aStack, ARMv7mStack, ARMv7rStack
from .i386 import X86Stack
from .loongarch import LoongArch64Stack
from .mips import MIPSBEStack, MIPSELStack
from .mips64 import MIPS64BEStack, MIPS64ELStack
from .ppc import PowerPC32Stack, PowerPC64Stack
from .riscv import RISCV64Stack
from .stack import *  # noqa: F401, F403
from .stack import __all__ as __stack__
from .xtensa import XTensaBEStack, XTensaELStack

__all__ = __stack__ + [
    "AArch64Stack",
    "AMD64Stack",
    "ARMv5tStack",
    "ARMv6mStack",
    "ARMv7mStack",
    "ARMv7rStack",
    "ARMv7aStack",
    "X86Stack",
    "LoongArch64Stack",
    "MIPSBEStack",
    "MIPSELStack",
    "MIPS64BEStack",
    "MIPS64ELStack",
    "PowerPC32Stack",
    "PowerPC64Stack",
    "RISCV64Stack",
    "XTensaBEStack",
    "XTensaELStack",
]



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import ArmELSysVModel


class ArmELSysVAbs(Abs, ArmELSysVModel):
    pass


class ArmELSysVLAbs(LAbs, ArmELSysVModel):
    pass


class ArmELSysVLLAbs(LLAbs, ArmELSysVModel):
    pass


class ArmELSysVAtof(Atof, ArmELSysVModel):
    pass


class ArmELSysVAtoi(Atoi, ArmELSysVModel):
    pass


class ArmELSysVAtol(Atol, ArmELSysVModel):
    pass


class ArmELSysVAtoll(Atoll, ArmELSysVModel):
    pass


class ArmELSysVCalloc(Calloc, ArmELSysVModel):
    pass


class ArmELSysVDiv(Div, ArmELSysVModel):
    pass


class ArmELSysVLDiv(LDiv, ArmELSysVModel):
    pass


class ArmELSysVLLDiv(LLDiv, ArmELSysVModel):
    pass


class ArmELSysVExit(Exit, ArmELSysVModel):
    pass


class ArmELSysVFree(Free, ArmELSysVModel):
    pass


class ArmELSysVMalloc(Malloc, ArmELSysVModel):
    pass


class ArmELSysVQSort(QSort, ArmELSysVModel):
    pass


class ArmELSysVRand(Rand, ArmELSysVModel):
    pass


class ArmELSysVRealloc(Realloc, ArmELSysVModel):
    pass


class ArmELSysVSrand(Srand, ArmELSysVModel):
    pass


class ArmELSysVAbort(Abort, ArmELSysVModel):
    pass


class ArmELSysVAtexit(Atexit, ArmELSysVModel):
    pass


class ArmELSysVGetenv(Getenv, ArmELSysVModel):
    pass


class ArmELSysVMblen(Mblen, ArmELSysVModel):
    pass


class ArmELSysVMbstowcs(Mbstowcs, ArmELSysVModel):
    pass


class ArmELSysVMbtowc(Mbtowc, ArmELSysVModel):
    pass


class ArmELSysVSystem(System, ArmELSysVModel):
    pass


class ArmELSysVWcstombs(Wcstombs, ArmELSysVModel):
    pass


class ArmELSysVWctomb(Wctomb, ArmELSysVModel):
    pass


__all__ = [
    "ArmELSysVAbs",
    "ArmELSysVLAbs",
    "ArmELSysVLLAbs",
    "ArmELSysVAtof",
    "ArmELSysVAtoi",
    "ArmELSysVAtol",
    "ArmELSysVAtoll",
    "ArmELSysVCalloc",
    "ArmELSysVDiv",
    "ArmELSysVLDiv",
    "ArmELSysVLLDiv",
    "ArmELSysVExit",
    "ArmELSysVFree",
    "ArmELSysVMalloc",
    "ArmELSysVQSort",
    "ArmELSysVRand",
    "ArmELSysVRealloc",
    "ArmELSysVSrand",
    "ArmELSysVAbort",
    "ArmELSysVAtexit",
    "ArmELSysVGetenv",
    "ArmELSysVMblen",
    "ArmELSysVMbstowcs",
    "ArmELSysVMbtowc",
    "ArmELSysVSystem",
    "ArmELSysVWcstombs",
    "ArmELSysVWctomb",
]



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import I386SysVModel


class I386SysVMemcpy(Memcpy, I386SysVModel):
    pass


class I386SysVMemmove(Memmove, I386SysVModel):
    pass


class I386SysVStrcpy(Strcpy, I386SysVModel):
    pass


class I386SysVStrncpy(Strncpy, I386SysVModel):
    pass


class I386SysVStrcat(Strcat, I386SysVModel):
    pass


class I386SysVStrncat(Strncat, I386SysVModel):
    pass


class I386SysVMemcmp(Memcmp, I386SysVModel):
    pass


class I386SysVStrncmp(Strncmp, I386SysVModel):
    pass


class I386SysVStrcmp(Strcmp, I386SysVModel):
    pass


class I386SysVStrcoll(Strcoll, I386SysVModel):
    pass


class I386SysVStrxfrm(Strxfrm, I386SysVModel):
    pass


class I386SysVMemchr(Memchr, I386SysVModel):
    pass


class I386SysVStrchr(Strchr, I386SysVModel):
    pass


class I386SysVStrcspn(Strcspn, I386SysVModel):
    pass


class I386SysVStrpbrk(Strpbrk, I386SysVModel):
    pass


class I386SysVStrrchr(Strrchr, I386SysVModel):
    pass


class I386SysVStrspn(Strspn, I386SysVModel):
    pass


class I386SysVStrstr(Strstr, I386SysVModel):
    pass


class I386SysVStrtok(Strtok, I386SysVModel):
    pass


class I386SysVMemset(Memset, I386SysVModel):
    pass


class I386SysVStrerror(Strerror, I386SysVModel):
    pass


class I386SysVStrlen(Strlen, I386SysVModel):
    pass


__all__ = [
    "I386SysVMemcpy",
    "I386SysVMemmove",
    "I386SysVStrcpy",
    "I386SysVStrncpy",
    "I386SysVStrcat",
    "I386SysVStrncat",
    "I386SysVMemcmp",
    "I386SysVStrncmp",
    "I386SysVStrcmp",
    "I386SysVStrcoll",
    "I386SysVStrxfrm",
    "I386SysVMemchr",
    "I386SysVStrchr",
    "I386SysVStrcspn",
    "I386SysVStrpbrk",
    "I386SysVStrrchr",
    "I386SysVStrspn",
    "I386SysVStrstr",
    "I386SysVStrtok",
    "I386SysVMemset",
    "I386SysVStrerror",
    "I386SysVStrlen",
]



from ....posix.libgen import Basename, Dirname
from ..systemv import PowerPCSysVModel


class PowerPCSysVBasename(Basename, PowerPCSysVModel):
    pass


class PowerPCSysVDirname(Dirname, PowerPCSysVModel):
    pass


__all__ = [
    "PowerPCSysVBasename",
    "PowerPCSysVDirname",
]



import copy
import hashlib
import logging
import random
import struct
import typing

import capstone

from .. import hinting, platforms, state
from ..exceptions import AnalysisRunError  # , EmulationBounds
from ..instructions import (
    BSIDMemoryReferenceOperand,
    Instruction,
    Operand,
    RegisterOperand,
)
from . import analysis
from .trace_execution import TraceExecution, TraceExecutionCBPoint

logger = logging.getLogger(__name__)


MIN_ACCEPTABLE_COLOR_INT = 0x20
BAD_COLOR = (2**64) - 1

Colors = typing.Dict[int, typing.Tuple[Operand, int, Instruction, int]]

Shad = typing.Dict[int, typing.Tuple[int, bool, int]]


def randomize_uninitialized(
    machine: state.Machine, seed: int = 123456, extra_regs: typing.List[str] = []
) -> state.Machine:
    """Consider all parts of the machine that can be written to (registers
    + memory regions). Write random values to any bytes in those machine
    parts which are currently uninitialized. So this only works if we
    have a way to tell if registers or memory have not been initialized.

    Randomize all general purpose regs (plus regs in list of
    extra_regs arg) that have not already been set.  Also, for any
    Heap, Stack, RawMemory or Memory regions, randomize any bytes that
    have not been set. This last part is to come since we don't
    currently have a way to tell which parts of memory have been
    written and which parts have not. Further, there are kinds of
    memory (Stack) which currently will break if we randomize all
    bytes.

    Note that, for a register, it is either set or not.  We can't tell
    if, say edx part of rdx has been set.

    """
    random.seed(seed)
    logger.setLevel(logging.DEBUG)

    # if logger.getEffectiveLevel() >= logging.DEBUG:
    m = hashlib.md5()
    platform = machine.get_platform()
    machine_copy = copy.deepcopy(machine)
    pdefs = platforms.defs.PlatformDef.for_platform(platform)
    reg_names = list(pdefs.registers.keys())
    reg_names.sort()

    def get_reg(machine, reg_name):
        cpu = machine.get_cpu()
        for x in cpu:
            if isinstance(x, state.Register):
                if x.name == reg_name:
                    return x
        return None

    for name in reg_names:
        if (name in pdefs.general_purpose_registers) or (name in extra_regs):
            reg = get_reg(machine_copy, name)
            if reg.get_content() is None:
                # this means reg is uninitialized; ok to randomize
                v = random.randint(0, (1 << (8 * reg.size)) - 1)
                reg.set(v)
                # logger.info(f"randomize_uninitialized setting {reg}")
                # if logger.getEffectiveLevel() >= logging.DEBUG:
                m.update(str(v).encode("utf-8"))

    for el in machine_copy:
        if isinstance(m, state.memory.Memory):
            if isinstance(m, state.memory.Executable):
                # nothing to randomize here
                continue
            # To come, I guess
            # we'd like to randomize anything uninitialized...
            pass
    # if logger.getEffectiveLevel() >= logging.DEBUG:
    logger.info(f"digest of changes made to machine: {m.hexdigest()}")

    logger.setLevel(logging.INFO)

    return machine_copy


class Colorizer(analysis.Analysis):
    """
    A simple kind of data flow analysis via tracking distinct values
    (colors) and employing instruction use/def analysis

    We run multiple micro-executions of the code starting from same
    entry. At the start of each, we randomize register values that
    have not already been initialized. We maintain a "colors" map from
    dynamic values to when/where we first observed them. This map is
    initially empty. Before emulating an instruction, we examine the
    values (registers and memory) it will read. If any are not in the
    colors map, that is the initial sighting of that value and we emit
    a hint to that effect and add a color to the map. If any color is
    already in the map, then that is a def-use flow from the time or
    place at which that value was first observed to this instruction.
    Similarly, after emulating an instruction, we examine every value
    written to a register or memory. If a value is not in the colors
    map, it is a new, computed result and we hint about its creation
    and add it to the map. If it is in the colors map, we do nothing
    since it just a copy.

    Whilst looking at reads and writes for instructions, we hint if any
    correspond to unavailable memory.

    Arguments:
        num_insns: The number of instructions to micro-execute.

    """

    name = "colorizer"
    description = "it's almost taint"
    version = "0.0.1"

    def __init__(
        self,
        *args,
        exec_id: int,
        num_insns: int = 200,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.exec_id = exec_id
        self.num_insns = num_insns
        self.colors: Colors = {}
        self.shadow_register: typing.Dict[str, Shad] = {}
        self.shadow_memory: Shad = {}
        # self.edge: typing.Dict[int, typing.Dict[int, typing.Tuple[str, int, int]]] = {}

    def _get_instr_at_pc(self, emu, pc: int) -> capstone.CsInsn:
        code = emu.read_memory(pc, 15)  # longest possible instruction
        if code is None:
            raise AnalysisRunError(
                "Unable to read next instruction out of emulator memory"
            )
        (insns, disas) = emu._disassemble(code, pc, 2)
        insn = insns[0]
        return insn

    def _operand_size(self, operand: Operand) -> int:
        if type(operand) is RegisterOperand:
            # return size of a reg based on its name
            return self.pdef.registers[operand.name].size
        elif type(operand) is BSIDMemoryReferenceOperand:
            # memory operand knows its size
            return operand.size
        return 0

    def run(self, machine: state.Machine) -> None:
        # collect hints for each microexecution, in a list of lists

        self.orig_machine = copy.deepcopy(machine)
        self.orig_cpu = self.orig_machine.get_cpu()
        self.platform = self.orig_cpu.platform
        self.pdef = platforms.PlatformDef.for_platform(self.platform)

        def check_rws(emu, pc, te, is_read):
            cs_insn = self._get_instr_at_pc(emu, pc)
            sw_insn = Instruction.from_capstone(cs_insn)
            if is_read:
                operand_list = sw_insn.reads
                lab = "read"
            else:
                operand_list = sw_insn.writes
                lab = "write"
            rws = []
            for operand in operand_list:
                logger.debug(f"pc={pc:x} {lab}_operand={operand}")
                if type(operand) is RegisterOperand and operand.name == "rflags":
                    continue
                sz = self._operand_size(operand)
                # print(f"operand={operand} sz={sz}")
                if type(operand) is BSIDMemoryReferenceOperand:
                    # if addr not mapped, discard this operand
                    a = operand.address(emu)
                    ar = (a, a + sz)
                    if not emu._is_address_range_mapped(ar):
                        continue
                conc = operand.concretize(emu)
                color = self._concrete_val_to_color(conc, sz)
                tup = (operand, conc, color, sz)
                rws.append(tup)
            rws.sort(key=lambda e: e[0].__repr__())
            if len(rws) == 0:
                return
            for rw in rws:
                (operand, conc, color, sz) = rw
                if color == BAD_COLOR:
                    pass
                else:
                    self._check_color(emu, is_read, rw, sw_insn, te.ic)
                # self.update_shadow(emu, pc, is_read, rw)
            if is_read:
                self.reads = rws

        def before_instruction_cb(emu, pc, te):
            check_rws(emu, pc, te, True)

        def after_instruction_cb(emu, pc, te):
            check_rws(emu, pc, te, False)

        self.colors = {}
        self.shadow_register = {}
        self.shadow_memory = {}
        traceA = TraceExecution(self.hinter, num_insns=self.num_insns)
        traceA.register_cb(
            TraceExecutionCBPoint.BEFORE_INSTRUCTION, before_instruction_cb
        )
        traceA.register_cb(
            TraceExecutionCBPoint.AFTER_INSTRUCTION, after_instruction_cb
        )
        traceA.run(machine)

        # NOTE: Please keep this code
        # if False:
        #     print("digraph{")
        #     print(" rankdir=LR")
        #     pc2nodeid = {}
        #     nodeid2pc = {}
        #     nodeids = set([])

        #     def add_pc(pc):
        #         if pc not in pc2nodeid:
        #             nodeid = f"node_{len(nodeids)}"
        #             nodeids.add(nodeid)
        #             pc2nodeid[pc] = nodeid
        #             nodeid2pc[nodeid] = pc

        #     for pc1 in self.edge.keys():
        #         add_pc(pc1)
        #         for pc2 in self.edge[pc1].keys():
        #             add_pc(pc2)
        #     for nodeid in nodeids:
        #         print(f'{nodeid} [label="0x{nodeid2pc[nodeid]:x}"]')
        #     for pc1 in self.edge.keys():
        #         for pc2 in self.edge[pc1].keys():
        #             (lab, conc, color) = self.edge[pc1][pc2]
        #             n1 = pc2nodeid[pc1]
        #             n2 = pc2nodeid[pc2]
        #             print(f'{n1} -> {n2} [label="{lab}"]')
        #     print("}")

    def _concrete_val_to_color(
        self, concrete_value: typing.Union[int, bytes, bytearray], size: int
    ) -> int:
        # this concrete value can be an int (if it came from a register)
        # or bytes (if it came from memory read)
        # we want these in a common format so that we can see them as colors
        the_bytes: bytes = b""
        if type(concrete_value) is int:
            if concrete_value < MIN_ACCEPTABLE_COLOR_INT:
                return BAD_COLOR
            the_bytes = concrete_value.to_bytes(size, byteorder="little")
        elif (type(concrete_value) is bytes) or (type(concrete_value) is bytearray):
            # assuming little-endian
            if (
                int.from_bytes(concrete_value, byteorder="little")
                < MIN_ACCEPTABLE_COLOR_INT
            ):
                return BAD_COLOR
            the_bytes = concrete_value
        else:
            assert 1 == 0
        # let's make color a number
        if size == 8:
            return struct.unpack("<Q", the_bytes)[0]
        if size == 4:
            return struct.unpack("<L", the_bytes)[0]
        if size == 2:
            return struct.unpack("<H", the_bytes)[0]
        assert size == 1
        return struct.unpack("<B", the_bytes)[0]

    def _add_color(
        self,
        color: int,
        operand: Operand,
        insn: Instruction,
        insn_num: int,
    ) -> None:
        self.colors[color] = (operand, insn_num, insn, 1 + len(self.colors))

    def _check_color(
        self,
        emu,
        is_read: bool,
        rw,  #: typing.Union[Operand, int, int],
        insn: Instruction,
        insn_num: int,
    ):
        (operand, conc, color, operand_size) = rw
        if color in self.colors.keys():
            # previously observed color
            if is_read:
                # read-flow: use of a previously recorded color value
                msg = "read-flow"
            else:
                # write of a previously seen value
                # ... its just a copy so no hint, right?
                msg = "write-copy"
            hint = self._dynamic_value_hint(
                emu,
                operand,
                operand_size,
                color,
                insn,
                is_read,
                False,
                insn_num,
                msg,
            )
            self.hinter.send(hint)
        else:
            # new color
            self._add_color(color, operand, insn, insn_num)
            if is_read:
                # read-def: use of a NOT previously recorded color value. As
                # long as the value is something reasonable, we'll record it as
                # a new color
                msg = "read-def"
            else:
                msg = "write-def"
            hint = self._dynamic_value_hint(
                emu,
                operand,
                operand_size,
                color,
                insn,
                is_read,
                True,
                insn_num,
                msg,
            )
            self.hinter.send(hint)

    def _dynamic_value_hint(
        self,
        emu,
        operand: Operand,
        size: int,
        color: int,
        insn: Instruction,
        is_use: bool,
        is_new: bool,
        insn_num: int,
        message: str,
    ):
        pc = insn.address
        if type(operand) is RegisterOperand:
            return hinting.DynamicRegisterValueHint(
                reg_name=operand.name,
                size=size,
                color=color,
                dynamic_value=color,
                use=is_use,
                new=is_new,
                pc=pc,
                instruction_num=insn_num,
                exec_id=self.exec_id,
                message=message,
            )
        elif type(operand) is BSIDMemoryReferenceOperand:
            base_name = "None"
            if operand.base is not None:
                base_name = operand.base
            index_name = "None"
            if operand.index is not None:
                index_name = operand.index
            return hinting.DynamicMemoryValueHint(
                address=operand.address(emu),
                base=base_name,
                index=index_name,
                scale=operand.scale,
                offset=operand.offset,
                color=color,
                dynamic_value=color,
                size=operand.size,
                use=is_use,
                new=is_new,
                pc=pc,
                instruction_num=insn_num,
                exec_id=self.exec_id,
                message=message,
            )
        else:
            assert 1 == 0

    # def update_shadow(self, emu, pc, is_read, rw):
    #     (operand, conc, color, operand_size) = rw

    #     if type(operand) is RegisterOperand:
    #         r = self.pdef.registers[operand.name]
    #         if type(r) is platforms.RegisterAliasDef:
    #             base_reg = r.parent
    #             start = r.offset
    #         else:
    #             base_reg = r.name
    #             start = 0
    #         if base_reg not in self.shadow_register:
    #             self.shadow_register[base_reg] = {}
    #         shad = self.shadow_register[base_reg]
    #         end = start + r.size
    #         lab = f"reg({r.name})"
    #     else:
    #         start = operand.address(emu)
    #         shad = self.shadow_memory
    #         end = start + operand.size
    #         lab = f"mem({start:x},{operand.size})"

    #     if is_read:
    #         # read. check labels on things we are reading to deduce flows
    #         fs = set([])
    #         for o in range(start, end):
    #             if o in shad:
    #                 (pc_from, is_read, conc_from) = shad[o]
    #                 if is_read:
    #                     f = f"{lab} r->r"
    #                 else:
    #                     f = f"{lab} w->r"
    #                 f += f" flow from pc={pc_from:x} to pc={pc:x} conc={conc} conc_from={conc_from} color={color}"
    #                 if f not in fs:
    #                     logger.info(f)
    #                     if pc_from not in self.edge:
    #                         self.edge[pc_from] = {}
    #                     if pc not in self.edge[pc_from]:
    #                         self.edge[pc_from][pc] = (lab, conc, color)
    #                     # self.edge.add((pc_from, lab, pc, conc, color))
    #                 fs.add(f)
    #     else:
    #         # write. we are overwriting things so no reason to check on bytes before doing so
    #         for o in range(start, end):
    #             shad[o] = (pc, is_read, conc)



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import PowerPCSysVModel


class PowerPCSysVMemcpy(Memcpy, PowerPCSysVModel):
    pass


class PowerPCSysVMemmove(Memmove, PowerPCSysVModel):
    pass


class PowerPCSysVStrcpy(Strcpy, PowerPCSysVModel):
    pass


class PowerPCSysVStrncpy(Strncpy, PowerPCSysVModel):
    pass


class PowerPCSysVStrcat(Strcat, PowerPCSysVModel):
    pass


class PowerPCSysVStrncat(Strncat, PowerPCSysVModel):
    pass


class PowerPCSysVMemcmp(Memcmp, PowerPCSysVModel):
    pass


class PowerPCSysVStrncmp(Strncmp, PowerPCSysVModel):
    pass


class PowerPCSysVStrcmp(Strcmp, PowerPCSysVModel):
    pass


class PowerPCSysVStrcoll(Strcoll, PowerPCSysVModel):
    pass


class PowerPCSysVStrxfrm(Strxfrm, PowerPCSysVModel):
    pass


class PowerPCSysVMemchr(Memchr, PowerPCSysVModel):
    pass


class PowerPCSysVStrchr(Strchr, PowerPCSysVModel):
    pass


class PowerPCSysVStrcspn(Strcspn, PowerPCSysVModel):
    pass


class PowerPCSysVStrpbrk(Strpbrk, PowerPCSysVModel):
    pass


class PowerPCSysVStrrchr(Strrchr, PowerPCSysVModel):
    pass


class PowerPCSysVStrspn(Strspn, PowerPCSysVModel):
    pass


class PowerPCSysVStrstr(Strstr, PowerPCSysVModel):
    pass


class PowerPCSysVStrtok(Strtok, PowerPCSysVModel):
    pass


class PowerPCSysVMemset(Memset, PowerPCSysVModel):
    pass


class PowerPCSysVStrerror(Strerror, PowerPCSysVModel):
    pass


class PowerPCSysVStrlen(Strlen, PowerPCSysVModel):
    pass


__all__ = [
    "PowerPCSysVMemcpy",
    "PowerPCSysVMemmove",
    "PowerPCSysVStrcpy",
    "PowerPCSysVStrncpy",
    "PowerPCSysVStrcat",
    "PowerPCSysVStrncat",
    "PowerPCSysVMemcmp",
    "PowerPCSysVStrncmp",
    "PowerPCSysVStrcmp",
    "PowerPCSysVStrcoll",
    "PowerPCSysVStrxfrm",
    "PowerPCSysVMemchr",
    "PowerPCSysVStrchr",
    "PowerPCSysVStrcspn",
    "PowerPCSysVStrpbrk",
    "PowerPCSysVStrrchr",
    "PowerPCSysVStrspn",
    "PowerPCSysVStrstr",
    "PowerPCSysVStrtok",
    "PowerPCSysVMemset",
    "PowerPCSysVStrerror",
    "PowerPCSysVStrlen",
]



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import PowerPCSysVModel


class PowerPCSysVTime(Time, PowerPCSysVModel):
    pass


class PowerPCSysVLocaltime(Localtime, PowerPCSysVModel):
    pass


class PowerPCSysVGmtime(Gmtime, PowerPCSysVModel):
    pass


class PowerPCSysVCtime(Ctime, PowerPCSysVModel):
    pass


class PowerPCSysVAsctime(Asctime, PowerPCSysVModel):
    pass


class PowerPCSysVStrftime(Strftime, PowerPCSysVModel):
    pass


class PowerPCSysVDifftime(Difftime, PowerPCSysVModel):
    pass


class PowerPCSysVMktime(Mktime, PowerPCSysVModel):
    pass


class PowerPCSysVClock(Clock, PowerPCSysVModel):
    pass


__all__ = [
    "PowerPCSysVTime",
    "PowerPCSysVLocaltime",
    "PowerPCSysVGmtime",
    "PowerPCSysVCtime",
    "PowerPCSysVAsctime",
    "PowerPCSysVStrftime",
    "PowerPCSysVDifftime",
    "PowerPCSysVMktime",
    "PowerPCSysVClock",
]



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import ArmELSysVModel


class ArmELSysVBsdSignal(BsdSignal, ArmELSysVModel):
    pass


class ArmELSysVKill(Kill, ArmELSysVModel):
    pass


class ArmELSysVKillpg(Killpg, ArmELSysVModel):
    pass


class ArmELSysVPthreadKill(PthreadKill, ArmELSysVModel):
    pass


class ArmELSysVPthreadSigmask(PthreadSigmask, ArmELSysVModel):
    pass


class ArmELSysVSigaction(Sigaction, ArmELSysVModel):
    pass


class ArmELSysVSigaddset(Sigaddset, ArmELSysVModel):
    pass


class ArmELSysVSigaltstack(Sigaltstack, ArmELSysVModel):
    pass


class ArmELSysVSigdelset(Sigdelset, ArmELSysVModel):
    pass


class ArmELSysVSigemptyset(Sigemptyset, ArmELSysVModel):
    pass


class ArmELSysVSigfillset(Sigfillset, ArmELSysVModel):
    pass


class ArmELSysVSighold(Sighold, ArmELSysVModel):
    pass


class ArmELSysVSigignore(Sigignore, ArmELSysVModel):
    pass


class ArmELSysVSiginterrupt(Siginterrupt, ArmELSysVModel):
    pass


class ArmELSysVSigismember(Sigismember, ArmELSysVModel):
    pass


class ArmELSysVSigpause(Sigpause, ArmELSysVModel):
    pass


class ArmELSysVSigpending(Sigpending, ArmELSysVModel):
    pass


class ArmELSysVSigprocmask(Sigprocmask, ArmELSysVModel):
    pass


class ArmELSysVSigqueue(Sigqueue, ArmELSysVModel):
    pass


class ArmELSysVSigrelse(Sigrelse, ArmELSysVModel):
    pass


class ArmELSysVSigset(Sigset, ArmELSysVModel):
    pass


class ArmELSysVSigsuspend(Sigsuspend, ArmELSysVModel):
    pass


class ArmELSysVSigtimedwait(Sigtimedwait, ArmELSysVModel):
    pass


class ArmELSysVSigwait(Sigwait, ArmELSysVModel):
    pass


class ArmELSysVSigwaitinfo(Sigwaitinfo, ArmELSysVModel):
    pass


__all__ = [
    "ArmELSysVBsdSignal",
    "ArmELSysVKill",
    "ArmELSysVKillpg",
    "ArmELSysVPthreadKill",
    "ArmELSysVPthreadSigmask",
    "ArmELSysVSigaction",
    "ArmELSysVSigaddset",
    "ArmELSysVSigaltstack",
    "ArmELSysVSigdelset",
    "ArmELSysVSigemptyset",
    "ArmELSysVSigfillset",
    "ArmELSysVSighold",
    "ArmELSysVSigignore",
    "ArmELSysVSiginterrupt",
    "ArmELSysVSigismember",
    "ArmELSysVSigpause",
    "ArmELSysVSigpending",
    "ArmELSysVSigprocmask",
    "ArmELSysVSigqueue",
    "ArmELSysVSigrelse",
    "ArmELSysVSigset",
    "ArmELSysVSigsuspend",
    "ArmELSysVSigtimedwait",
    "ArmELSysVSigwait",
    "ArmELSysVSigwaitinfo",
]



from ....c99.signal import Raise, Signal
from ..systemv import MIPSELSysVModel


class MIPSELSysVRaise(Raise, MIPSELSysVModel):
    pass


class MIPSELSysVSignal(Signal, MIPSELSysVModel):
    pass


__all__ = [
    "MIPSELSysVRaise",
    "MIPSELSysVSignal",
]



import typing

import capstone

from ..platforms import Architecture, Byteorder
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef

# PowerPC's register naming convention is extrmely annoying:
# it doesn't have one.
#
# Both the general-purpose registers
# and floating-point registers are solely referenced by number,
# thus making it maddening to tell which one you're talking about.
#
# Also, registers do have special uses within the ABI,
# but aside from sp and bp, they're not really written down.


class PowerPCPlatformDef(PlatformDef):
    byteorder = Byteorder.BIG

    capstone_arch = capstone.CS_ARCH_PPC

    # PowerPC conditional branches are an abject pain.
    # There are 172 possible ways to conditionally branch to an address,
    # and a similar number of ways to branch to LR or CTR.
    # Not all of these have mnemonics, but the ISA doesn't actually specify which.

    # Branch to a label
    _branch_to_label = {
        # Conditional branch
        "bc",
        "beq",
        "bne",
        "blt",
        "ble",
        "bgt",
        "bge",
        "bso",
        "bns",
        # Decrement ctr and branch if non-zero
        # NOTE: This can also include a conditional test
        "bdnz",
        # Decrement ctr and branch if zero
        # NOTE: This can also include a conditional test
        "bdz",
    }
    # Handle "branch-and-link" and "absolute address" variants
    _branch_and_link_to_label = set(map(lambda x: f"{x}l", _branch_to_label))
    _branch_absolute_to_label = set(map(lambda x: f"{x}a", _branch_to_label))
    _branch_and_link_absolute_to_label = set(
        map(lambda x: f"{x}a", _branch_and_link_to_label)
    )

    # Handle branch predictor suggestions
    _branch_to_label_neutral = (
        _branch_to_label
        | _branch_and_link_to_label
        | _branch_absolute_to_label
        | _branch_and_link_absolute_to_label
    )
    _branch_to_label_likely = set(map(lambda x: f"{x}+", _branch_to_label_neutral))
    _branch_to_label_unlikely = set(map(lambda x: f"{x}-", _branch_to_label_neutral))

    # Collect branch to label
    _branch_to_label_all = (
        _branch_to_label_neutral | _branch_to_label_likely | _branch_to_label_unlikely
    )

    # Branch to LR looks the same as branch to label, but with an "lr" suffix after the condition code
    _branch_to_lr = set(map(lambda x: f"{x}lr", _branch_to_label))
    _branch_and_link_to_lr = set(map(lambda x: f"{x}l", _branch_to_lr))

    # Handle branch predictor suggestions
    _branch_to_lr_neutral = _branch_to_lr | _branch_and_link_to_lr
    _branch_to_lr_likely = set(map(lambda x: f"{x}+", _branch_to_lr_neutral))
    _branch_to_lr_unlikely = set(map(lambda x: f"{x}-", _branch_to_lr_neutral))

    # Collect branch to LR
    _branch_to_lr_all = (
        _branch_to_lr_neutral | _branch_to_lr_likely | _branch_to_lr_unlikely
    )

    # Branch to CTR looks the same as branch to label, but with a "ctr" suffix after the condition code.
    _branch_to_ctr = set(map(lambda x: f"{x}ctr", _branch_to_label))
    # There are no mnemonics for using ctr as a counter and branch target at once.
    _branch_to_ctr -= {"bdnzctr", "bdzctr"}
    _branch_and_link_to_ctr = set(map(lambda x: f"{x}l", _branch_to_ctr))

    # Handle branch predictor suggestions
    _branch_to_ctr_neutral = _branch_to_ctr | _branch_and_link_to_ctr
    _branch_to_ctr_likely = set(map(lambda x: f"{x}+", _branch_to_ctr_neutral))
    _branch_to_ctr_unlikely = set(map(lambda x: f"{x}-", _branch_to_ctr_neutral))

    # Collect branch to CTR
    _branch_to_ctr_all = (
        _branch_to_ctr_neutral | _branch_to_ctr_likely | _branch_to_lr_unlikely
    )

    # Finally, collect all conditional branch mnemonics
    conditional_branch_mnemonics = (
        _branch_to_label_all | _branch_to_lr_all | _branch_to_ctr_all
    )

    compare_mnemonics = {
        # Compare registers
        "cmpd",
        "cmpw",
        # Compare immediate
        "cmpi",
        "cmpdi",
        "cmpwi",
        # Compare logical immediate
        "cmpli",
        "cmpldi",
        "cmplwi",
        # Compare
        "cmpl",
        "cmpld",
        "cmplw",
    }

    pc_register = "pc"
    sp_register = "sp"

    # Special registers
    # - r0 and r2 are reserved for the assembler.
    # - r1 is the stack pointer
    # - r31 is the base (frame) pointer
    general_purpose_registers = [f"r{i}" for i in range(3, 31)]

    @property
    def registers(self) -> typing.Dict[str, RegisterDef]:
        return self._registers

    def __init__(self):
        self._registers = {
            # *** General Purpose Registers ***
            # NOTE: Used expressive names for GPRs and FPRs.
            # gasm just refers to GPRs and FPRS by number.
            # They use the same numbers; it's very annoying.
            "r0": RegisterDef(name="r0", size=self.address_size),
            # NOTE: GPR 1 is also the stack pointer.
            "r1": RegisterDef(name="r1", size=self.address_size),
            "sp": RegisterAliasDef("sp", parent="r1", size=self.address_size, offset=0),
            "r2": RegisterDef(name="r2", size=self.address_size),
            "r3": RegisterDef(name="r3", size=self.address_size),
            "r4": RegisterDef(name="r4", size=self.address_size),
            "r5": RegisterDef(name="r5", size=self.address_size),
            "r6": RegisterDef(name="r6", size=self.address_size),
            "r7": RegisterDef(name="r7", size=self.address_size),
            "r8": RegisterDef(name="r8", size=self.address_size),
            "r9": RegisterDef(name="r9", size=self.address_size),
            "r10": RegisterDef(name="r10", size=self.address_size),
            "r11": RegisterDef(name="r11", size=self.address_size),
            "r12": RegisterDef(name="r12", size=self.address_size),
            "r13": RegisterDef(name="r13", size=self.address_size),
            "r14": RegisterDef(name="r14", size=self.address_size),
            "r15": RegisterDef(name="r15", size=self.address_size),
            "r16": RegisterDef(name="r16", size=self.address_size),
            "r17": RegisterDef(name="r17", size=self.address_size),
            "r18": RegisterDef(name="r18", size=self.address_size),
            "r19": RegisterDef(name="r19", size=self.address_size),
            "r20": RegisterDef(name="r20", size=self.address_size),
            "r21": RegisterDef(name="r21", size=self.address_size),
            "r22": RegisterDef(name="r22", size=self.address_size),
            "r23": RegisterDef(name="r23", size=self.address_size),
            "r24": RegisterDef(name="r24", size=self.address_size),
            "r25": RegisterDef(name="r25", size=self.address_size),
            "r26": RegisterDef(name="r26", size=self.address_size),
            "r27": RegisterDef(name="r27", size=self.address_size),
            "r28": RegisterDef(name="r28", size=self.address_size),
            "r29": RegisterDef(name="r29", size=self.address_size),
            "r30": RegisterDef(name="r30", size=self.address_size),
            # NOTE: GPR 31 is also the base pointer
            "r31": RegisterDef(name="r31", size=self.address_size),
            "bp": RegisterAliasDef(
                name="bp", parent="r31", size=self.address_size, offset=0
            ),
            # Floating Point Registers
            # Always 8 bytes, regardless of self.address_size.
            "f0": RegisterDef(name="f0", size=8),
            "f1": RegisterDef(name="f1", size=8),
            "f2": RegisterDef(name="f2", size=8),
            "f3": RegisterDef(name="f3", size=8),
            "f4": RegisterDef(name="f4", size=8),
            "f5": RegisterDef(name="f5", size=8),
            "f6": RegisterDef(name="f6", size=8),
            "f7": RegisterDef(name="f7", size=8),
            "f8": RegisterDef(name="f8", size=8),
            "f9": RegisterDef(name="f9", size=8),
            "f10": RegisterDef(name="f10", size=8),
            "f11": RegisterDef(name="f11", size=8),
            "f12": RegisterDef(name="f12", size=8),
            "f13": RegisterDef(name="f13", size=8),
            "f14": RegisterDef(name="f14", size=8),
            "f15": RegisterDef(name="f15", size=8),
            "f16": RegisterDef(name="f16", size=8),
            "f17": RegisterDef(name="f17", size=8),
            "f18": RegisterDef(name="f18", size=8),
            "f19": RegisterDef(name="f19", size=8),
            "f20": RegisterDef(name="f20", size=8),
            "f21": RegisterDef(name="f21", size=8),
            "f22": RegisterDef(name="f22", size=8),
            "f23": RegisterDef(name="f23", size=8),
            "f24": RegisterDef(name="f24", size=8),
            "f25": RegisterDef(name="f25", size=8),
            "f26": RegisterDef(name="f26", size=8),
            "f27": RegisterDef(name="f27", size=8),
            "f28": RegisterDef(name="f28", size=8),
            "f29": RegisterDef(name="f29", size=8),
            "f30": RegisterDef(name="f30", size=8),
            "f31": RegisterDef(name="f31", size=8),
            # *** Pointer Registers ***
            # Program Counter.
            # Not really a register; nothing can access it directly
            "pc": RegisterDef(name="pc", size=self.address_size),
            # Link Register
            "lr": RegisterDef(name="lr", size=self.address_size),
            # Counter Register
            # Acts either as a loop index, or a branch target register
            # Only `ctr` and `lr` can act as branch targets.
            "ctr": RegisterDef(name="ctr", size=self.address_size),
            # *** Condition Registers ***
            # The actual condition register `cr` is a single 32-bit register,
            # but it's broken into eight 4-bit fields which are accessed separately.
            #
            # Certain operations use specific registers by default,
            # but some tests can specify a destination register.
            "cr0": RegisterDef(name="cr0", size=1),  # Integer condition bits
            "cr1": RegisterDef(name="cr1", size=1),  # Floating point condition bits
            "cr2": RegisterDef(name="cr2", size=1),
            "cr3": RegisterDef(name="cr3", size=1),
            "cr4": RegisterDef(name="cr4", size=1),
            "cr5": RegisterDef(name="cr5", size=1),
            "cr6": RegisterDef(name="cr6", size=1),
            "cr7": RegisterDef(name="cr7", size=1),
            # Integer Exception Register
            "xer": RegisterDef(name="xer", size=4),
            # Floating Point Status and Control Register
            "fpscr": RegisterDef(name="fpscr", size=4),
            # TODO: This only focuses on the user-facing registrers.
            # ppc has a huge number of privileged registers.
            # Extend this as needed.
        }


class PowerPC32(PowerPCPlatformDef):
    architecture = Architecture.POWERPC32

    address_size = 4
    capstone_mode = capstone.CS_MODE_32 | capstone.CS_MODE_BIG_ENDIAN


class PowerPC64(PowerPCPlatformDef):
    architecture = Architecture.POWERPC64

    address_size = 8
    capstone_mode = capstone.CS_MODE_64 | capstone.CS_MODE_BIG_ENDIAN



import logging
import typing

from ...emulators import AngrEmulator
from ...exceptions import EmulationStop
from ...platforms import Platform
from ...state import Machine
from ..underlays import AnalysisUnderlay

log = logging.getLogger(__name__)


class ForcedExecutionUnderlay(AnalysisUnderlay):
    """Forced execution analysis underlay

    This allows you to emulate arbitrary program slices
    by forcing the emulator to visit specific instructions,
    ignoring the normal program control flow.

    This isn't a complete analysis;
    instead, it's a dummy base class that makes it easy
    to support forced execution mode by implementing
    the complex logic as an overlay,
    and building the actual analysis as a combination
    of overlay and underlay.
    See field detection to see what I mean.

    NOTE: This is not compatible with all architectures.
    The architecture needs to support single-stepping;
    delay slot architectures such as MIPS can't
    be single-stepped by angr.

    Arguments:
        trace:      The list of program counter addresses you want to visit

    """

    def __init__(self, trace: typing.List[typing.Dict[str, int]]):
        self.trace: typing.List[typing.Dict[str, int]] = trace

    def execute(self):
        try:
            for regs in self.trace:
                for reg, val in regs.items():
                    self.emulator.write_register_content(reg, val)
                self.emulator.step()
        except EmulationStop:
            pass


class ForcedExecution(ForcedExecutionUnderlay):
    """Forced execution using angr

    This allows you to emulate arbitrary program slices
    by forcing the emulator to visit specific instructions,
    ignoring the normal program control flow.

    NOTE: This is not compatible with all architectures.
    The architecture needs to support single-stepping;
    delay slot architectures such as MIPS can't
    be single-stepped by angr.

    Arguments:
        platform:   The platform you want to emulate
        trace:      The list of program counter addresses you want to visit
    """

    name = "forced-execution"
    description = "Forced execution using angr"
    version = "0.0.1"

    def __init__(self, platform: Platform, trace: typing.List[typing.Dict[str, int]]):
        super().__init__(trace)
        self.platform: Platform = platform
        self.emulator = AngrEmulator(platform)
        self.emulator.enable_linear()

    def run(self, machine: Machine):
        emulator = self.emulator
        if not isinstance(emulator, AngrEmulator):
            raise TypeError("Impossible!")
        machine.apply(emulator)
        emulator.initialize()
        self.execute()
        for s in emulator.mgr.active:
            log.info(s.solver.constraints)
            s.registers.pp(log.info)



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import MIPSELSysVModel


class MIPSELSysVBsdSignal(BsdSignal, MIPSELSysVModel):
    pass


class MIPSELSysVKill(Kill, MIPSELSysVModel):
    pass


class MIPSELSysVKillpg(Killpg, MIPSELSysVModel):
    pass


class MIPSELSysVPthreadKill(PthreadKill, MIPSELSysVModel):
    pass


class MIPSELSysVPthreadSigmask(PthreadSigmask, MIPSELSysVModel):
    pass


class MIPSELSysVSigaction(Sigaction, MIPSELSysVModel):
    pass


class MIPSELSysVSigaddset(Sigaddset, MIPSELSysVModel):
    pass


class MIPSELSysVSigaltstack(Sigaltstack, MIPSELSysVModel):
    pass


class MIPSELSysVSigdelset(Sigdelset, MIPSELSysVModel):
    pass


class MIPSELSysVSigemptyset(Sigemptyset, MIPSELSysVModel):
    pass


class MIPSELSysVSigfillset(Sigfillset, MIPSELSysVModel):
    pass


class MIPSELSysVSighold(Sighold, MIPSELSysVModel):
    pass


class MIPSELSysVSigignore(Sigignore, MIPSELSysVModel):
    pass


class MIPSELSysVSiginterrupt(Siginterrupt, MIPSELSysVModel):
    pass


class MIPSELSysVSigismember(Sigismember, MIPSELSysVModel):
    pass


class MIPSELSysVSigpause(Sigpause, MIPSELSysVModel):
    pass


class MIPSELSysVSigpending(Sigpending, MIPSELSysVModel):
    pass


class MIPSELSysVSigprocmask(Sigprocmask, MIPSELSysVModel):
    pass


class MIPSELSysVSigqueue(Sigqueue, MIPSELSysVModel):
    pass


class MIPSELSysVSigrelse(Sigrelse, MIPSELSysVModel):
    pass


class MIPSELSysVSigset(Sigset, MIPSELSysVModel):
    pass


class MIPSELSysVSigsuspend(Sigsuspend, MIPSELSysVModel):
    pass


class MIPSELSysVSigtimedwait(Sigtimedwait, MIPSELSysVModel):
    pass


class MIPSELSysVSigwait(Sigwait, MIPSELSysVModel):
    pass


class MIPSELSysVSigwaitinfo(Sigwaitinfo, MIPSELSysVModel):
    pass


__all__ = [
    "MIPSELSysVBsdSignal",
    "MIPSELSysVKill",
    "MIPSELSysVKillpg",
    "MIPSELSysVPthreadKill",
    "MIPSELSysVPthreadSigmask",
    "MIPSELSysVSigaction",
    "MIPSELSysVSigaddset",
    "MIPSELSysVSigaltstack",
    "MIPSELSysVSigdelset",
    "MIPSELSysVSigemptyset",
    "MIPSELSysVSigfillset",
    "MIPSELSysVSighold",
    "MIPSELSysVSigignore",
    "MIPSELSysVSiginterrupt",
    "MIPSELSysVSigismember",
    "MIPSELSysVSigpause",
    "MIPSELSysVSigpending",
    "MIPSELSysVSigprocmask",
    "MIPSELSysVSigqueue",
    "MIPSELSysVSigrelse",
    "MIPSELSysVSigset",
    "MIPSELSysVSigsuspend",
    "MIPSELSysVSigtimedwait",
    "MIPSELSysVSigwait",
    "MIPSELSysVSigwaitinfo",
]



import logging
import random
import typing

import claripy

from .... import emulators, exceptions
from ...memory.heap import Heap
from ..cstd import ArgumentType, CStdModel
from .utils import _emu_strlen

logger = logging.getLogger("__name__")


class Abort(CStdModel):
    name = "abort"

    # void abort(void);
    argument_types = []
    return_type = ArgumentType.VOID

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise exceptions.EmulationStop("Called abort()")


class Abs(CStdModel):
    name = "abs"

    # int abs(int val);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.INT

    @property
    def sign_mask(self):
        return self._int_sign_mask

    @property
    def inv_mask(self):
        return self._int_inv_mask

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        val = self.get_arg1(emulator)

        assert isinstance(val, int)

        if val & self.sign_mask:
            val = ((val ^ self.inv_mask) + 1) & self.inv_mask

        self.set_return_value(emulator, val)


class LAbs(Abs):
    name = "labs"

    # long labs(long x);
    argument_types = [ArgumentType.LONG]
    return_type = ArgumentType.LONG

    @property
    def sign_mask(self):
        return self._long_sign_mask

    @property
    def inv_mask(self):
        return self._long_inv_mask


class LLAbs(Abs):
    name = "llabs"

    # long long llabs(long long x);
    argument_types = [ArgumentType.LONGLONG]
    return_type = ArgumentType.LONGLONG

    @property
    def sign_mask(self):
        return self._long_long_sign_mask

    @property
    def inv_mask(self):
        return self._long_long_inv_mask


class Atexit(CStdModel):
    name = "atexit"

    # NOTE: In glibc binaries, relocate atexit against __cxa_atexit
    # atexit is a statically-linked helper that calls __cxa_atexit

    # void atexit(void);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # This will not actually result in an exit handler getting registered.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        self.set_return_value(emulator, 0)


class Atof(CStdModel):
    name = "atof"

    # float atof(const char *str);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.FLOAT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # TODO: Support other locales for atof
        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        n = _emu_strlen(emulator, ptr)

        data = emulator.read_memory(ptr, n)
        text = data.decode("utf-8")

        # This is a bit tricky.  Python is much less accepting than C.
        text = text.strip()
        found_dot = False
        for i in range(0, len(text)):
            if text[i].isnumeric():
                continue
            elif text[i] == ".":
                if found_dot:
                    text = text[0:i]
                    break
                else:
                    found_dot = True
            else:
                text = text[0:i]
                break
        if len(text) == 0:
            text = "0"

        self.set_return_value(emulator, float(text))


class Atoi(CStdModel):
    name = "atoi"

    # int atoi(const char *str);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    size_mask = 0xFFFFFFFF

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # TODO: Support other locales for atoi
        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        n = _emu_strlen(emulator, ptr)

        data = emulator.read_memory(ptr, n)
        text = data.decode("utf-8").strip()
        print(f"Converting {text} ({len(text)})")

        for i in range(0, len(text)):
            if not text[i].isnumeric() and text[i] != "-":
                text = text[0:i]
                break

        if len(text) == 0:
            # No valid number
            self.set_return_value(emulator, 0)
            return

        # TODO: Not entirely sure if this is how truncation will work.
        newval = int(text) & self.size_mask
        self.set_return_value(emulator, newval)


class Atol(Atoi):
    name = "atol"

    # long atoll(const char *str);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.LONG

    size_mask = 0xFFFFFFFFFFFFFFFF


class Atoll(Atoi):
    name = "atoll"

    # long long atoll(const char *str);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.LONGLONG

    size_mask = 0xFFFFFFFFFFFFFFFF


class Bsearch(CStdModel):
    name = "bsearch"

    # void *bsearch(const void *key, const void *base,
    #               size_t nitems, size_t size,
    #               int (*compar)(const void *, const void *));
    argument_types = [
        ArgumentType.POINTER,
        ArgumentType.POINTER,
        ArgumentType.SIZE_T,
        ArgumentType.SIZE_T,
        ArgumentType.POINTER,
    ]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # Not easily possible; need to call a comparator function.
        raise NotImplementedError(
            "bsearch uses a function pointer; not sure how to model"
        )


class Calloc(CStdModel):
    name = "calloc"

    # void *calloc(size_t amount, size_t size);
    argument_types = [ArgumentType.SIZE_T, ArgumentType.SIZE_T]
    return_type = ArgumentType.POINTER

    def __init__(self, address: int):
        super().__init__(address)
        # Use the same heap model the harness used.
        # NOTE: This will get cloned on a deep copy.
        self.heap: typing.Optional[Heap] = None

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        if self.heap is None:
            raise exceptions.ConfigurationError(
                "calloc needs a heap; please assign self.heap"
            )

        amt = self.get_arg1(emulator)
        size = self.get_arg2(emulator)

        assert isinstance(amt, int)
        assert isinstance(size, int)

        data = b"\0" * amt * size

        res = self.heap.allocate_bytes(data, None)
        # This is calloc; zero out the memory
        emulator.write_memory(res, data)

        self.set_return_value(emulator, res)


class Div(CStdModel):
    name = "div"

    # div_t result(int dividend, int divisor);
    argument_types = [ArgumentType.INT, ArgumentType.INT]
    # FIXME: Figure out how different platforms return structs by value

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # div_t result(int dividend, int divisor);
        raise NotImplementedError()


class LDiv(Div):
    name = "ldiv"

    # ldiv_t result(long dividend, long divisor);
    argument_types = [ArgumentType.LONG, ArgumentType.LONG]
    # FIXME: Figure out how different platforms return structs by value


class LLDiv(Div):
    name = "lldiv"

    # lldiv_t result(long long dividend, long long divisor);
    argument_types = [ArgumentType.LONGLONG, ArgumentType.LONGLONG]
    # FIXME: Figure out how different platforms return structs by value


class Exit(CStdModel):
    name = "exit"

    # void exit(int code);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.VOID

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        raise exceptions.EmulationStop("Called exit()")


class Free(CStdModel):
    name = "free"

    # void free(void *ptr);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.VOID

    def __init__(self, address: int):
        super().__init__(address)
        # Use the same heap model the harness used.
        # NOTE: This will get cloned on a deep copy.
        self.heap: typing.Optional[Heap] = None

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        if self.heap is None:
            raise exceptions.ConfigurationError(
                "malloc needs a heap; please assign self.heap"
            )

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        self.heap.free(ptr)


class Getenv(CStdModel):
    name = "getenv"

    # char *getenv(char *name);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    # We don't have a model of envp,
    # so this will always return NULL
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        size = _emu_strlen(emulator, ptr)
        data = emulator.read_memory(ptr, size)
        name = data.decode("utf-8")

        logger.info(f"getenv({name});")
        self.set_return_value(emulator, 0)


class Malloc(CStdModel):
    name = "malloc"

    # void *malloc(size_t size);
    argument_types = [ArgumentType.SIZE_T]
    return_type = ArgumentType.POINTER

    def __init__(self, address: int):
        super().__init__(address)
        # Use the same heap model the harness used.
        # NOTE: This will get cloned on a deep copy.
        self.heap: typing.Optional[Heap] = None

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        if self.heap is None:
            raise exceptions.ConfigurationError(
                "malloc needs a heap; please assign self.heap"
            )

        size = self.get_arg1(emulator)

        assert isinstance(size, int)

        res = self.heap.allocate_bytes(b"\0" * size, None)

        self.set_return_value(emulator, res)


class Mblen(CStdModel):
    name = "mblen"

    # int mblen(char *str, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # Depends the locale.
        raise NotImplementedError()


class Mbstowcs(CStdModel):
    name = "mbstowcs"

    # size_t mbstowcs(schar_t *pwcs, char *str, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # Depends the locale.
        raise NotImplementedError()


class Mbtowc(CStdModel):
    name = "mbtowc"

    # size_t mbtowc(wchar_t *pwcs, char *str, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # Depends the locale.
        raise NotImplementedError()


class QSort(CStdModel):
    name = "qsort"

    # void qsort(void *arr, size_t amount, size_t size, int (*compare)(const void *, const void *));
    argument_types = [
        ArgumentType.POINTER,
        ArgumentType.SIZE_T,
        ArgumentType.SIZE_T,
        ArgumentType.POINTER,
    ]
    return_type = ArgumentType.VOID

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # Not easily possible; need to call a comparator function.
        raise NotImplementedError(
            "qsort uses a function pointer; not sure how to model"
        )


class Rand(CStdModel):
    name = "rand"

    # int rand(void);
    argument_types = []
    return_type = ArgumentType.INT

    rand = random.Random()

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # TODO: Rand is easy to do simply, harder to do right.
        # If someone is relying on srand/rand to produce a specific sequence,
        # this won't behave correctly.
        val = self.rand.randint(0, 2147483647)
        self.set_return_value(emulator, val)


class Realloc(CStdModel):
    name = "realloc"

    # void *realloc(void *ptr, size_t size);
    argument_types = [ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.POINTER

    def __init__(self, address: int):
        super().__init__(address)
        # Use the same heap model the harness used.
        # NOTE: This will get cloned on a deep copy.
        self.heap: typing.Optional[Heap] = None

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        if self.heap is None:
            raise exceptions.ConfigurationError(
                "realloc needs a heap; please assign self.heap"
            )

        ptr = self.get_arg1(emulator)
        size = self.get_arg2(emulator)

        assert isinstance(ptr, int)
        assert isinstance(size, int)

        logger.warning(f"REALLOC {hex(ptr)}, {size}")

        if ptr == 0:
            res = self.heap.allocate_bytes(b"\0" * size, None)
        elif ptr - self.heap.address not in self.heap:
            raise exceptions.EmulationError(
                f"Attempted to realloc {hex(ptr)}, which was not malloc'd on this heap"
            )
        else:
            oldsize = self.heap[ptr - self.heap.address].get_size()
            data: typing.Union[bytes, claripy.ast.bv.BV]
            try:
                data = emulator.read_memory(ptr, oldsize)
            except exceptions.SymbolicValueError:
                data = emulator.read_memory_symbolic(ptr, oldsize)

            self.heap.free(ptr)

            res = self.heap.allocate_bytes(b"\0" * size, None)
            emulator.write_memory(res, data)

        self.set_return_value(emulator, res)


class Srand(CStdModel):
    name = "srand"

    # void srand(int seed);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.VOID

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        seed = self.get_arg1(emulator)
        Rand.rand.seed(a=seed)


class System(CStdModel):
    name = "system"

    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        size = _emu_strlen(emulator, ptr)
        data = emulator.read_memory(ptr, size)
        cmd = data.decode("utf-8")

        logger.info(f"system({cmd});")
        self.set_return_value(emulator, 0)


class Wcstombs(CStdModel):
    name = "wctombs"

    # size_t wctombs(char *str, wchar_t *pwcs, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # Depends the locale.
        raise NotImplementedError()


class Wctomb(CStdModel):
    name = "wctomb"

    # int wctomb(char *str, wchar_t wchar);
    argument_types = [ArgumentType.POINTER, ArgumentType.UINT]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # Depends the locale.
        raise NotImplementedError()


__all__ = [
    "Abs",
    "LAbs",
    "LLAbs",
    "Abort",
    "Atexit",
    "Atof",
    "Atoi",
    "Atol",
    "Atoll",
    "Calloc",
    "Div",
    "LDiv",
    "LLDiv",
    "Exit",
    "Free",
    "Getenv",
    "Malloc",
    "Mblen",
    "Mbstowcs",
    "Mbtowc",
    "QSort",
    "Rand",
    "Realloc",
    "Srand",
    "System",
    "Wcstombs",
    "Wctomb",
]



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import MIPSSysVModel


class MIPSSysVMemcpy(Memcpy, MIPSSysVModel):
    pass


class MIPSSysVMemmove(Memmove, MIPSSysVModel):
    pass


class MIPSSysVStrcpy(Strcpy, MIPSSysVModel):
    pass


class MIPSSysVStrncpy(Strncpy, MIPSSysVModel):
    pass


class MIPSSysVStrcat(Strcat, MIPSSysVModel):
    pass


class MIPSSysVStrncat(Strncat, MIPSSysVModel):
    pass


class MIPSSysVMemcmp(Memcmp, MIPSSysVModel):
    pass


class MIPSSysVStrncmp(Strncmp, MIPSSysVModel):
    pass


class MIPSSysVStrcmp(Strcmp, MIPSSysVModel):
    pass


class MIPSSysVStrcoll(Strcoll, MIPSSysVModel):
    pass


class MIPSSysVStrxfrm(Strxfrm, MIPSSysVModel):
    pass


class MIPSSysVMemchr(Memchr, MIPSSysVModel):
    pass


class MIPSSysVStrchr(Strchr, MIPSSysVModel):
    pass


class MIPSSysVStrcspn(Strcspn, MIPSSysVModel):
    pass


class MIPSSysVStrpbrk(Strpbrk, MIPSSysVModel):
    pass


class MIPSSysVStrrchr(Strrchr, MIPSSysVModel):
    pass


class MIPSSysVStrspn(Strspn, MIPSSysVModel):
    pass


class MIPSSysVStrstr(Strstr, MIPSSysVModel):
    pass


class MIPSSysVStrtok(Strtok, MIPSSysVModel):
    pass


class MIPSSysVMemset(Memset, MIPSSysVModel):
    pass


class MIPSSysVStrerror(Strerror, MIPSSysVModel):
    pass


class MIPSSysVStrlen(Strlen, MIPSSysVModel):
    pass


__all__ = [
    "MIPSSysVMemcpy",
    "MIPSSysVMemmove",
    "MIPSSysVStrcpy",
    "MIPSSysVStrncpy",
    "MIPSSysVStrcat",
    "MIPSSysVStrncat",
    "MIPSSysVMemcmp",
    "MIPSSysVStrncmp",
    "MIPSSysVStrcmp",
    "MIPSSysVStrcoll",
    "MIPSSysVStrxfrm",
    "MIPSSysVMemchr",
    "MIPSSysVStrchr",
    "MIPSSysVStrcspn",
    "MIPSSysVStrpbrk",
    "MIPSSysVStrrchr",
    "MIPSSysVStrspn",
    "MIPSSysVStrstr",
    "MIPSSysVStrtok",
    "MIPSSysVMemset",
    "MIPSSysVStrerror",
    "MIPSSysVStrlen",
]



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import AMD64SysVModel


class AMD64SysVBsdSignal(BsdSignal, AMD64SysVModel):
    pass


class AMD64SysVKill(Kill, AMD64SysVModel):
    pass


class AMD64SysVKillpg(Killpg, AMD64SysVModel):
    pass


class AMD64SysVPthreadKill(PthreadKill, AMD64SysVModel):
    pass


class AMD64SysVPthreadSigmask(PthreadSigmask, AMD64SysVModel):
    pass


class AMD64SysVSigaction(Sigaction, AMD64SysVModel):
    pass


class AMD64SysVSigaddset(Sigaddset, AMD64SysVModel):
    pass


class AMD64SysVSigaltstack(Sigaltstack, AMD64SysVModel):
    pass


class AMD64SysVSigdelset(Sigdelset, AMD64SysVModel):
    pass


class AMD64SysVSigemptyset(Sigemptyset, AMD64SysVModel):
    pass


class AMD64SysVSigfillset(Sigfillset, AMD64SysVModel):
    pass


class AMD64SysVSighold(Sighold, AMD64SysVModel):
    pass


class AMD64SysVSigignore(Sigignore, AMD64SysVModel):
    pass


class AMD64SysVSiginterrupt(Siginterrupt, AMD64SysVModel):
    pass


class AMD64SysVSigismember(Sigismember, AMD64SysVModel):
    pass


class AMD64SysVSigpause(Sigpause, AMD64SysVModel):
    pass


class AMD64SysVSigpending(Sigpending, AMD64SysVModel):
    pass


class AMD64SysVSigprocmask(Sigprocmask, AMD64SysVModel):
    pass


class AMD64SysVSigqueue(Sigqueue, AMD64SysVModel):
    pass


class AMD64SysVSigrelse(Sigrelse, AMD64SysVModel):
    pass


class AMD64SysVSigset(Sigset, AMD64SysVModel):
    pass


class AMD64SysVSigsuspend(Sigsuspend, AMD64SysVModel):
    pass


class AMD64SysVSigtimedwait(Sigtimedwait, AMD64SysVModel):
    pass


class AMD64SysVSigwait(Sigwait, AMD64SysVModel):
    pass


class AMD64SysVSigwaitinfo(Sigwaitinfo, AMD64SysVModel):
    pass


__all__ = [
    "AMD64SysVBsdSignal",
    "AMD64SysVKill",
    "AMD64SysVKillpg",
    "AMD64SysVPthreadKill",
    "AMD64SysVPthreadSigmask",
    "AMD64SysVSigaction",
    "AMD64SysVSigaddset",
    "AMD64SysVSigaltstack",
    "AMD64SysVSigdelset",
    "AMD64SysVSigemptyset",
    "AMD64SysVSigfillset",
    "AMD64SysVSighold",
    "AMD64SysVSigignore",
    "AMD64SysVSiginterrupt",
    "AMD64SysVSigismember",
    "AMD64SysVSigpause",
    "AMD64SysVSigpending",
    "AMD64SysVSigprocmask",
    "AMD64SysVSigqueue",
    "AMD64SysVSigrelse",
    "AMD64SysVSigset",
    "AMD64SysVSigsuspend",
    "AMD64SysVSigtimedwait",
    "AMD64SysVSigwait",
    "AMD64SysVSigwaitinfo",
]



from ....c99.signal import Raise, Signal
from ..systemv import ArmHFSysVModel


class ArmHFSysVRaise(Raise, ArmHFSysVModel):
    pass


class ArmHFSysVSignal(Signal, ArmHFSysVModel):
    pass


__all__ = [
    "ArmHFSysVRaise",
    "ArmHFSysVSignal",
]



import capstone

from ..platforms import Architecture, Byteorder
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef


class AArch64(PlatformDef):
    architecture = Architecture.AARCH64
    byteorder = Byteorder.LITTLE

    address_size = 8
    capstone_arch = capstone.CS_ARCH_ARM64
    capstone_mode = capstone.CS_MODE_ARM

    pc_register = "pc"
    sp_register = "sp"

    # NOTE: aarch64 has unconditional conditions.
    # "al" is always true, and "nv" is never true.
    conditional_branch_mnemonics = {
        # Conditional branches
        "b.eq",
        "b.ne",
        "b.cs",
        "b.cc",
        "b.mi",
        "b.pl",
        "b.vs",
        "b.vc",
        "b.hi",
        "b.ls",
        "b.ge",
        "b.lt",
        "b.gt",
        "b.le",
        # Consistent (unlikely?) conditional branches
        "bc.eq",
        "bc.ne",
        "bc.cs",
        "bc.cc",
        "bc.mi",
        "bc.pl",
        "bc.vs",
        "bc.vc",
        "bc.hi",
        "bc.ls",
        "bc.ge",
        "bc.lt",
        "bc.gt",
        "bc.le",
        # Compare and conditional branch
        # TODO: How to handle this?
        "cb.eq",
        "cb.ne",
        "cb.cs",
        "cb.cc",
        "cb.mi",
        "cb.pl",
        "cb.vs",
        "cb.vc",
        "cb.hi",
        "cb.ls",
        "cb.ge",
        "cb.lt",
        "cb.gt",
        "cb.le",
        # Extra mnemonics for cb opcodes
        "cbge",
        "cbhs",
        "cble",
        "cblo",
        "cbls",
        "cblt",
        "cbnz",
        "cbz",
        # Compare bytes and conditional branch
        "cbb.eq",
        "cbb.ne",
        "cbb.cs",
        "cbb.cc",
        "cbb.mi",
        "cbb.pl",
        "cbb.vs",
        "cbb.vc",
        "cbb.hi",
        "cbb.ls",
        "cbb.ge",
        "cbb.lt",
        "cbb.gt",
        "cbb.le",
        # Extra mnemonics for cbb opcodes
        "cbble",
        "cbblo",
        "cbbls",
        "cbblt",
        # Compare half-words and conditional branch
        "cbh.eq",
        "cbh.ne",
        "cbh.cs",
        "cbh.cc",
        "cbh.mi",
        "cbh.pl",
        "cbh.vs",
        "cbh.vc",
        "cbh.hi",
        "cbh.ls",
        "cbh.ge",
        "cbh.lt",
        "cbh.gt",
        "cbh.le",
        # Extra mnemonics for cbh opcodes
        "cbhle",
        "cbhlo",
        "cbhls",
        "cbhlt",
    }

    # TODO: Should arithmetic operations that impact flags be compares?
    compare_mnemonics = {
        # Integer comparison
        # AArch64 supports positive and negative variants of comparisons
        "cmn",
        "cmp",
        "tst",
        # Conditional comparison
        # This performs a comparison if the condition flags pass the desired test,
        # otherwise it sets the flags to an immediate.
        # NOTE: This takes a condition type, but it's encoded as an operand.
        "ccmn",
        "ccmp",
        # Compare with tag
        # Performs 56-bit comparison, likely to support tagged pointers
        "cmpp",
        # Vector comparisons
        "cmeq",
        "cmge",
        "cmgt",
        "cmhi",
        "cmhs",
        "cmle",
        "cmlt",
        "cmtst",
        # Floating-point comparisons
        "fcmeq",
        "fcmge",
        "fcmgt",
        "fcmle",
        "fcmlt",
        "fcmp",
        "fcmpe",
        # Floating-point conditional comparisons
        # NOTE: This takes a condition type, but it's encoded as an operand
        "fccmp",
        "fccmpe",
    }

    # Special registers:
    # x29: frame pointer
    # x30: link register
    # x31: stack pointer or zero, depending on instruction
    general_purpose_registers = [f"x{i}" for i in range(0, 29)]

    registers = {
        # *** General Purpose Registers ***
        "x0": RegisterDef(name="x0", size=8),
        "w0": RegisterAliasDef(name="w0", parent="x0", size=4, offset=0),
        "x1": RegisterDef(name="x1", size=8),
        "w1": RegisterAliasDef(name="w1", parent="x1", size=4, offset=0),
        "x2": RegisterDef(name="x2", size=8),
        "w2": RegisterAliasDef(name="w2", parent="x2", size=4, offset=0),
        "x3": RegisterDef(name="x3", size=8),
        "w3": RegisterAliasDef(name="w3", parent="x3", size=4, offset=0),
        "x4": RegisterDef(name="x4", size=8),
        "w4": RegisterAliasDef(name="w4", parent="x4", size=4, offset=0),
        "x5": RegisterDef(name="x5", size=8),
        "w5": RegisterAliasDef(name="w5", parent="x5", size=4, offset=0),
        "x6": RegisterDef(name="x6", size=8),
        "w6": RegisterAliasDef(name="w6", parent="x6", size=4, offset=0),
        "x7": RegisterDef(name="x7", size=8),
        "w7": RegisterAliasDef(name="w7", parent="x7", size=4, offset=0),
        "x8": RegisterDef(name="x8", size=8),
        "w8": RegisterAliasDef(name="w8", parent="x8", size=4, offset=0),
        "x9": RegisterDef(name="x9", size=8),
        "w9": RegisterAliasDef(name="w9", parent="x9", size=4, offset=0),
        "x10": RegisterDef(name="x10", size=8),
        "w10": RegisterAliasDef(name="w10", parent="x10", size=4, offset=0),
        "x11": RegisterDef(name="x11", size=8),
        "w11": RegisterAliasDef(name="w11", parent="x10", size=4, offset=0),
        "x12": RegisterDef(name="x12", size=8),
        "w12": RegisterAliasDef(name="w12", parent="x10", size=4, offset=0),
        "x13": RegisterDef(name="x13", size=8),
        "w13": RegisterAliasDef(name="w13", parent="x13", size=4, offset=0),
        "x14": RegisterDef(name="x14", size=8),
        "w14": RegisterAliasDef(name="w14", parent="x14", size=4, offset=0),
        "x15": RegisterDef(name="x15", size=8),
        "w15": RegisterAliasDef(name="w15", parent="x15", size=4, offset=0),
        "x16": RegisterDef(name="x16", size=8),
        "w16": RegisterAliasDef(name="w16", parent="x16", size=4, offset=0),
        "x17": RegisterDef(name="x17", size=8),
        "w17": RegisterAliasDef(name="w17", parent="x17", size=4, offset=0),
        "x18": RegisterDef(name="x18", size=8),
        "w18": RegisterAliasDef(name="w18", parent="x18", size=4, offset=0),
        "x19": RegisterDef(name="x19", size=8),
        "w19": RegisterAliasDef(name="w19", parent="x19", size=4, offset=0),
        "x20": RegisterDef(name="x20", size=8),
        "w20": RegisterAliasDef(name="w20", parent="x20", size=4, offset=0),
        "x21": RegisterDef(name="x21", size=8),
        "w21": RegisterAliasDef(name="w21", parent="x20", size=4, offset=0),
        "x22": RegisterDef(name="x22", size=8),
        "w22": RegisterAliasDef(name="w22", parent="x20", size=4, offset=0),
        "x23": RegisterDef(name="x23", size=8),
        "w23": RegisterAliasDef(name="w23", parent="x23", size=4, offset=0),
        "x24": RegisterDef(name="x24", size=8),
        "w24": RegisterAliasDef(name="w24", parent="x24", size=4, offset=0),
        "x25": RegisterDef(name="x25", size=8),
        "w25": RegisterAliasDef(name="w25", parent="x25", size=4, offset=0),
        "x26": RegisterDef(name="x26", size=8),
        "w26": RegisterAliasDef(name="w26", parent="x26", size=4, offset=0),
        "x27": RegisterDef(name="x27", size=8),
        "w27": RegisterAliasDef(name="w27", parent="x27", size=4, offset=0),
        "x28": RegisterDef(name="x28", size=8),
        "w28": RegisterAliasDef(name="w28", parent="x28", size=4, offset=0),
        "x29": RegisterDef(name="x29", size=8),
        "w29": RegisterAliasDef(name="w29", parent="x29", size=4, offset=0),
        "x30": RegisterDef(name="x30", size=8),
        "w30": RegisterAliasDef(name="w30", parent="x30", size=4, offset=0),
        # *** Program Counter ***
        "pc": RegisterDef(name="pc", size=8),
        # *** Stack Pointer ***
        "sp": RegisterDef(name="sp", size=8),
        "wsp": RegisterAliasDef(name="wsp", parent="sp", size=4, offset=0),
        # *** Frame Pointer ***
        "fp": RegisterAliasDef(name="fp", parent="x29", size=8, offset=0),
        # *** Link Register ***
        "lr": RegisterAliasDef(name="lr", parent="x30", size=8, offset=0),
        # *** Zero Register ***
        "xzr": RegisterDef(name="xzr", size=8),
        "wzr": RegisterAliasDef(name="wzr", parent="xzr", size=4, offset=0),
        # *** System Control Registers ***
        # NOTE: "_elX" indicates that only exception level X or greater can access this register.
        # NOTE: This list is far from complete; it only covers what Unicorn supports
        # Condition Code Register
        "fpcr": RegisterDef(name="fpcr", size=8),
        # Floating Point Status Register
        "fpsr": RegisterDef(name="fpsr", size=8),
        # Banked stack pointers for exception handlers
        "sp_el0": RegisterDef(name="sp_el0", size=8),
        "sp_el1": RegisterDef(name="sp_el1", size=8),
        "sp_el2": RegisterDef(name="sp_el2", size=8),
        "sp_el3": RegisterDef(name="sp_el3", size=8),
        # Banked link registers for exception handlers
        # NOTE: Unicorn thinks there's an elr_el0; according to docs, it doesn't exist
        "elr_el1": RegisterDef(name="elr_el1", size=8),
        "elr_el2": RegisterDef(name="elr_el2", size=8),
        "elr_el3": RegisterDef(name="elr_el3", size=8),
        # Banked exception syndrome registers for exception handlers
        # NOTE: Unicorn thinks there's a far_el0; according to docs, it doesn't exist
        "far_el1": RegisterDef(name="far_el1", size=8),
        "far_el2": RegisterDef(name="far_el2", size=8),
        "far_el3": RegisterDef(name="far_el3", size=8),
        # Banked vector base address registers for exception handlers
        # NOTE: vbar_el0 and vbar_el1 are aliases for each other.
        # Since vbar_el0 doesn't exist in angr, vbar_el1 has to be the "real" copy.
        "vbar_el1": RegisterDef(name="vbar_el1", size=8),
        "vbar_el0": RegisterAliasDef(
            name="vbar_el0", parent="vbar_el1", size=8, offset=0
        ),
        "vbar_el2": RegisterDef(name="vbar_el2", size=8),
        "vbar_el3": RegisterDef(name="vbar_el3", size=8),
        # Coprocessor access control register
        "cpacr_el1": RegisterDef(name="cpacr_el1", size=8),
        # Memory Attribute Indirection Register
        "mair_el1": RegisterDef(name="mair_el1", size=8),
        # Physical Address Register
        "par_el1": RegisterDef(name="par_el1", size=8),
        # Translation Table Zero Base Register
        "ttbr0_el1": RegisterDef(name="ttbr0_el1", size=8),
        # Translation Table One Base Register
        "ttbr1_el1": RegisterDef(name="ttbr1_el1", size=8),
        # Thread ID Register
        # NOTE: According to docs, there should be an el2 and el3 copy, too.
        "tpidr_el0": RegisterDef(name="tpidr_el0", size=8),
        "tpidr_el1": RegisterDef(name="tpidr_el1", size=8),
        # Userspace-visible Thread ID register
        "tpidrro_el0": RegisterDef(name="tpidrro_el0", size=8),
        # *** Floating Point Registers ***
        # Scalar Floating Point Registers
        "q0": RegisterDef(name="q0", size=16),
        "d0": RegisterAliasDef(name="d0", parent="q0", size=8, offset=0),
        "s0": RegisterAliasDef(name="s0", parent="q0", size=4, offset=0),
        "h0": RegisterAliasDef(name="h0", parent="q0", size=2, offset=0),
        "b0": RegisterAliasDef(name="b0", parent="q0", size=1, offset=0),
        "q1": RegisterDef(name="q1", size=16),
        "d1": RegisterAliasDef(name="d1", parent="q1", size=8, offset=0),
        "s1": RegisterAliasDef(name="s1", parent="q1", size=4, offset=0),
        "h1": RegisterAliasDef(name="h1", parent="q1", size=2, offset=0),
        "b1": RegisterAliasDef(name="b1", parent="q1", size=1, offset=0),
        "q2": RegisterDef(name="q2", size=16),
        "d2": RegisterAliasDef(name="d2", parent="q2", size=8, offset=0),
        "s2": RegisterAliasDef(name="s2", parent="q2", size=4, offset=0),
        "h2": RegisterAliasDef(name="h2", parent="q2", size=2, offset=0),
        "b2": RegisterAliasDef(name="b2", parent="q2", size=1, offset=0),
        "q3": RegisterDef(name="q3", size=16),
        "d3": RegisterAliasDef(name="d3", parent="q3", size=8, offset=0),
        "s3": RegisterAliasDef(name="s3", parent="q3", size=4, offset=0),
        "h3": RegisterAliasDef(name="h3", parent="q3", size=2, offset=0),
        "b3": RegisterAliasDef(name="b3", parent="q3", size=1, offset=0),
        "q4": RegisterDef(name="q4", size=16),
        "d4": RegisterAliasDef(name="d4", parent="q4", size=8, offset=0),
        "s4": RegisterAliasDef(name="s4", parent="q4", size=4, offset=0),
        "h4": RegisterAliasDef(name="h4", parent="q4", size=2, offset=0),
        "b4": RegisterAliasDef(name="b4", parent="q4", size=1, offset=0),
        "q5": RegisterDef(name="q5", size=16),
        "d5": RegisterAliasDef(name="d5", parent="q5", size=8, offset=0),
        "s5": RegisterAliasDef(name="s5", parent="q5", size=4, offset=0),
        "h5": RegisterAliasDef(name="h5", parent="q5", size=2, offset=0),
        "b5": RegisterAliasDef(name="b5", parent="q5", size=1, offset=0),
        "q6": RegisterDef(name="q6", size=16),
        "d6": RegisterAliasDef(name="d6", parent="q6", size=8, offset=0),
        "s6": RegisterAliasDef(name="s6", parent="q6", size=4, offset=0),
        "h6": RegisterAliasDef(name="h6", parent="q6", size=2, offset=0),
        "b6": RegisterAliasDef(name="b6", parent="q6", size=1, offset=0),
        "q7": RegisterDef(name="q7", size=16),
        "d7": RegisterAliasDef(name="d7", parent="q7", size=8, offset=0),
        "s7": RegisterAliasDef(name="s7", parent="q7", size=4, offset=0),
        "h7": RegisterAliasDef(name="h7", parent="q7", size=2, offset=0),
        "b7": RegisterAliasDef(name="b7", parent="q7", size=1, offset=0),
        "q8": RegisterDef(name="q8", size=16),
        "d8": RegisterAliasDef(name="d8", parent="q8", size=8, offset=0),
        "s8": RegisterAliasDef(name="s8", parent="q8", size=4, offset=0),
        "h8": RegisterAliasDef(name="h8", parent="q8", size=2, offset=0),
        "b8": RegisterAliasDef(name="b8", parent="q8", size=1, offset=0),
        "q9": RegisterDef(name="q9", size=16),
        "d9": RegisterAliasDef(name="d9", parent="q9", size=8, offset=0),
        "s9": RegisterAliasDef(name="s9", parent="q9", size=4, offset=0),
        "h9": RegisterAliasDef(name="h9", parent="q9", size=2, offset=0),
        "b9": RegisterAliasDef(name="b9", parent="q9", size=1, offset=0),
        "q10": RegisterDef(name="q10", size=16),
        "d10": RegisterAliasDef(name="d10", parent="q10", size=8, offset=0),
        "s10": RegisterAliasDef(name="s10", parent="q10", size=4, offset=0),
        "h10": RegisterAliasDef(name="h10", parent="q10", size=2, offset=0),
        "b10": RegisterAliasDef(name="b10", parent="q10", size=1, offset=0),
        "q11": RegisterDef(name="q11", size=16),
        "d11": RegisterAliasDef(name="d11", parent="q11", size=8, offset=0),
        "s11": RegisterAliasDef(name="s11", parent="q11", size=4, offset=0),
        "h11": RegisterAliasDef(name="h11", parent="q11", size=2, offset=0),
        "b11": RegisterAliasDef(name="b11", parent="q11", size=1, offset=0),
        "q12": RegisterDef(name="q12", size=16),
        "d12": RegisterAliasDef(name="d12", parent="q12", size=8, offset=0),
        "s12": RegisterAliasDef(name="s12", parent="q12", size=4, offset=0),
        "h12": RegisterAliasDef(name="h12", parent="q12", size=2, offset=0),
        "b12": RegisterAliasDef(name="b12", parent="q12", size=1, offset=0),
        "q13": RegisterDef(name="q13", size=16),
        "d13": RegisterAliasDef(name="d13", parent="q13", size=8, offset=0),
        "s13": RegisterAliasDef(name="s13", parent="q13", size=4, offset=0),
        "h13": RegisterAliasDef(name="h13", parent="q13", size=2, offset=0),
        "b13": RegisterAliasDef(name="b13", parent="q13", size=1, offset=0),
        "q14": RegisterDef(name="q14", size=16),
        "d14": RegisterAliasDef(name="d14", parent="q14", size=8, offset=0),
        "s14": RegisterAliasDef(name="s14", parent="q14", size=4, offset=0),
        "h14": RegisterAliasDef(name="h14", parent="q14", size=2, offset=0),
        "b14": RegisterAliasDef(name="b14", parent="q14", size=1, offset=0),
        "q15": RegisterDef(name="q15", size=16),
        "d15": RegisterAliasDef(name="d15", parent="q15", size=8, offset=0),
        "s15": RegisterAliasDef(name="s15", parent="q15", size=4, offset=0),
        "h15": RegisterAliasDef(name="h15", parent="q15", size=2, offset=0),
        "b15": RegisterAliasDef(name="b15", parent="q15", size=1, offset=0),
        "q16": RegisterDef(name="q16", size=16),
        "d16": RegisterAliasDef(name="d16", parent="q16", size=8, offset=0),
        "s16": RegisterAliasDef(name="s16", parent="q16", size=4, offset=0),
        "h16": RegisterAliasDef(name="h16", parent="q16", size=2, offset=0),
        "b16": RegisterAliasDef(name="b16", parent="q16", size=1, offset=0),
        "q17": RegisterDef(name="q17", size=16),
        "d17": RegisterAliasDef(name="d17", parent="q17", size=8, offset=0),
        "s17": RegisterAliasDef(name="s17", parent="q17", size=4, offset=0),
        "h17": RegisterAliasDef(name="h17", parent="q17", size=2, offset=0),
        "b17": RegisterAliasDef(name="b17", parent="q17", size=1, offset=0),
        "q18": RegisterDef(name="q18", size=16),
        "d18": RegisterAliasDef(name="d18", parent="q18", size=8, offset=0),
        "s18": RegisterAliasDef(name="s18", parent="q18", size=4, offset=0),
        "h18": RegisterAliasDef(name="h18", parent="q18", size=2, offset=0),
        "b18": RegisterAliasDef(name="b18", parent="q18", size=1, offset=0),
        "q19": RegisterDef(name="q19", size=16),
        "d19": RegisterAliasDef(name="d19", parent="q19", size=8, offset=0),
        "s19": RegisterAliasDef(name="s19", parent="q19", size=4, offset=0),
        "h19": RegisterAliasDef(name="h19", parent="q19", size=2, offset=0),
        "b19": RegisterAliasDef(name="b19", parent="q19", size=1, offset=0),
        "q20": RegisterDef(name="q20", size=16),
        "d20": RegisterAliasDef(name="d20", parent="q20", size=8, offset=0),
        "s20": RegisterAliasDef(name="s20", parent="q20", size=4, offset=0),
        "h20": RegisterAliasDef(name="h20", parent="q20", size=2, offset=0),
        "b20": RegisterAliasDef(name="b20", parent="q20", size=1, offset=0),
        "q21": RegisterDef(name="q21", size=16),
        "d21": RegisterAliasDef(name="d21", parent="q21", size=8, offset=0),
        "s21": RegisterAliasDef(name="s21", parent="q21", size=4, offset=0),
        "h21": RegisterAliasDef(name="h21", parent="q21", size=2, offset=0),
        "b21": RegisterAliasDef(name="b21", parent="q21", size=1, offset=0),
        "q22": RegisterDef(name="q22", size=16),
        "d22": RegisterAliasDef(name="d22", parent="q22", size=8, offset=0),
        "s22": RegisterAliasDef(name="s22", parent="q22", size=4, offset=0),
        "h22": RegisterAliasDef(name="h22", parent="q22", size=2, offset=0),
        "b22": RegisterAliasDef(name="b22", parent="q22", size=1, offset=0),
        "q23": RegisterDef(name="q23", size=16),
        "d23": RegisterAliasDef(name="d23", parent="q23", size=8, offset=0),
        "s23": RegisterAliasDef(name="s23", parent="q23", size=4, offset=0),
        "h23": RegisterAliasDef(name="h23", parent="q23", size=2, offset=0),
        "b23": RegisterAliasDef(name="b23", parent="q23", size=1, offset=0),
        "q24": RegisterDef(name="q24", size=16),
        "d24": RegisterAliasDef(name="d24", parent="q24", size=8, offset=0),
        "s24": RegisterAliasDef(name="s24", parent="q24", size=4, offset=0),
        "h24": RegisterAliasDef(name="h24", parent="q24", size=2, offset=0),
        "b24": RegisterAliasDef(name="b24", parent="q24", size=1, offset=0),
        "q25": RegisterDef(name="q25", size=16),
        "d25": RegisterAliasDef(name="d25", parent="q25", size=8, offset=0),
        "s25": RegisterAliasDef(name="s25", parent="q25", size=4, offset=0),
        "h25": RegisterAliasDef(name="h25", parent="q25", size=2, offset=0),
        "b25": RegisterAliasDef(name="b25", parent="q25", size=1, offset=0),
        "q26": RegisterDef(name="q26", size=16),
        "d26": RegisterAliasDef(name="d26", parent="q26", size=8, offset=0),
        "s26": RegisterAliasDef(name="s26", parent="q26", size=4, offset=0),
        "h26": RegisterAliasDef(name="h26", parent="q26", size=2, offset=0),
        "b26": RegisterAliasDef(name="b26", parent="q26", size=1, offset=0),
        "q27": RegisterDef(name="q27", size=16),
        "d27": RegisterAliasDef(name="d27", parent="q27", size=8, offset=0),
        "s27": RegisterAliasDef(name="s27", parent="q27", size=4, offset=0),
        "h27": RegisterAliasDef(name="h27", parent="q27", size=2, offset=0),
        "b27": RegisterAliasDef(name="b27", parent="q27", size=1, offset=0),
        "q28": RegisterDef(name="q28", size=16),
        "d28": RegisterAliasDef(name="d28", parent="q28", size=8, offset=0),
        "s28": RegisterAliasDef(name="s28", parent="q28", size=4, offset=0),
        "h28": RegisterAliasDef(name="h28", parent="q28", size=2, offset=0),
        "b28": RegisterAliasDef(name="b28", parent="q28", size=1, offset=0),
        "q29": RegisterDef(name="q29", size=16),
        "d29": RegisterAliasDef(name="d29", parent="q29", size=8, offset=0),
        "s29": RegisterAliasDef(name="s29", parent="q29", size=4, offset=0),
        "h29": RegisterAliasDef(name="h29", parent="q29", size=2, offset=0),
        "b29": RegisterAliasDef(name="b29", parent="q29", size=1, offset=0),
        "q30": RegisterDef(name="q30", size=16),
        "d30": RegisterAliasDef(name="d30", parent="q30", size=8, offset=0),
        "s30": RegisterAliasDef(name="s30", parent="q30", size=4, offset=0),
        "h30": RegisterAliasDef(name="h30", parent="q30", size=2, offset=0),
        "b30": RegisterAliasDef(name="b30", parent="q30", size=1, offset=0),
        "q31": RegisterDef(name="q31", size=16),
        "d31": RegisterAliasDef(name="d31", parent="q31", size=8, offset=0),
        "s31": RegisterAliasDef(name="s31", parent="q31", size=4, offset=0),
        "h31": RegisterAliasDef(name="h31", parent="q31", size=2, offset=0),
        "b31": RegisterAliasDef(name="b31", parent="q31", size=1, offset=0),
        # Vector registers
        # TODO: Figure out how to model these
    }
